import os
import uuid
from io import BytesIO

import pandas as pd
from gcloud.aio.storage import Storage
from loguru import logger

from bb_integrations_lib.models.pipeline_structs import BolExportResults
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.shared.model import RawData, FileConfigRawData


class ArchiveGCSStep(Step):
    def __init__(
            self,
            gcloud_storage: Storage,
            bucket_path: str,
            field_sep: str = ",",
            content_type: str = "",
            error_on_exists: bool = False,
            *args,
            **kwargs
    ) -> None:
        """
        Archive a file (RawData or BolExportResults) to Google Cloud Storage.
        :param gcloud_storage: A gcloud-aio-storage client.
        :param bucket_path: The bucket and optional directory to upload the file to.
          Example: "my-bucket" or "my-bucket/my-dir/my-dir-2". If a FileConfigRawData is passed to the step, the
          source_system will be appended to the directory, before the file name. (bucket/prefix/source_system/file_name)
        :param field_sep: Field separator when uploading BolExportResults.
        :param content_type: Optional - explicitly specify the content type of the file.
        :param error_on_exists: Whether to raise an exception if the file exists.
        """
        super().__init__(*args, **kwargs)

        self.storage = gcloud_storage
        self.bucket = bucket_path
        self.prefix = ""
        self.field_sep = field_sep
        self.content_type = content_type
        self.error_on_exists = error_on_exists

        if "/" in bucket_path:
            [self.bucket, self.prefix] = bucket_path.split("/", maxsplit=1)

    def describe(self):
        return "Archiving file in GCS"

    async def execute(self, i: RawData | FileConfigRawData | BolExportResults) -> RawData:
        def file_path(base_name: str):
            if isinstance(i, FileConfigRawData):
                return os.path.join(self.prefix, i.file_config.source_system, base_name)
            else:
                return os.path.join(self.prefix, base_name)

        # Will be used to change the name if a duplicate is detected.
        # Don't want to change it directly on the original object because that str will get used by later steps
        file_name = i.file_name

        if isinstance(i, BolExportResults):
            df = pd.DataFrame.from_records(i.orders)
            csv_text = df.to_csv(index=False, sep=self.field_sep)
            # Default to text/csv for BolExportResults if we don't have an explicit content type
            self.content_type = "text/csv" if not self.content_type else self.content_type
            contents = csv_text.encode("utf-8")
        else:
            contents = i.data

        if isinstance(contents, BytesIO):
            contents = contents.getvalue()

        if await self.storage.get_bucket(self.bucket).blob_exists(file_path(file_name)):
            if self.error_on_exists:
                raise Exception(f"File '{self.bucket}/{file_path(file_name)}' already exists")
            old_file_name = file_name
            file_name = f"DUPLICATE_{uuid.uuid4()}_{file_name}"
            logger.debug(
                f"Blob named '{old_file_name}' already exists; archiving '{file_name}' to GCS (backup attempt)")
        else:
            logger.debug(f"Archiving '{file_name}' to GCS")
        await self.storage.upload(self.bucket, file_path(file_name), contents or '', content_type=self.content_type)

        return i



import time
from loguru import logger
from typing import TypeVar, Callable, Dict, Tuple, Any
import asyncio

T = TypeVar('T')

class CustomTTLCache:
    def __init__(self, verbose: bool = False):
        self.caches: dict = {}
        self.verbose = verbose

    def _is_expired(self, timestamp: float) -> bool:
        return time.time() > timestamp

    def ttl_cache(self, seconds: int,  cache_key: str):
        def decorator(func: Callable[..., T]) -> Callable[..., T]:
            def wrapper(*args, **kwargs):
                if cache_key in self.caches:
                    value, expires_at = self.caches[cache_key]
                    if time.time() < expires_at:
                        if self.verbose:
                            logger.info(f"Cache HIT for {cache_key}")
                        return value
                    else:
                        if self.verbose:
                            logger.info(f"Cache EXPIRED for {cache_key}")
                        del self.caches[cache_key]
                else:
                    if self.verbose:
                        logger.info(f"Cache MISS for {cache_key}")
                result = func(*args, **kwargs)
                expires_at = time.time() + seconds
                self.caches[cache_key] = (result, expires_at)
                if self.verbose:
                    logger.info(
                        f"Cached result for {cache_key} (expires at {time.strftime('%H:%M:%S', time.localtime(expires_at))})")
                return result
            return wrapper
        return decorator


class CustomAsyncTTLCache:
    def __init__(self, verbose: bool = False):
        self.caches: Dict[str, Tuple[Any, float]] = {}
        self.verbose = verbose
        self.locks: Dict[str, asyncio.Lock] = {}

    def _is_expired(self, timestamp: float) -> bool:
        return time.time() > timestamp

    async def get_or_set(self, cache_key: str, seconds: int, fetch_func, *args, **kwargs):
        if cache_key not in self.locks:
            self.locks[cache_key] = asyncio.Lock()
        async with self.locks[cache_key]:
            if cache_key in self.caches:
                value, expires_at = self.caches[cache_key]
                if time.time() < expires_at:
                    if self.verbose:
                        logger.info(f"Cache HIT for {cache_key}")
                    return value
                else:
                    if self.verbose:
                        logger.info(f"Cache EXPIRED for {cache_key}")
                    del self.caches[cache_key]
            else:
                if self.verbose:
                    logger.info(f"Cache MISS for {cache_key}")
            result = await fetch_func(*args, **kwargs)
            expires_at = time.time() + seconds
            self.caches[cache_key] = (result, expires_at)
            if self.verbose:
                logger.info(
                    f"Cached result for {cache_key} (expires at {time.strftime('%H:%M:%S', time.localtime(expires_at))})")
            return result


from datetime import datetime
from typing import Literal, Optional, Any, override, List

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass

from bb_integrations_lib.models.rita.issue import IssueBase


@dataclass
class CrossroadsError(Exception):
    issue: IssueBase = Field(..., description="An issue to be saved to logs")


@dataclass
class CrossroadsMappingError(Exception):
    """Custom exception type that """
    friendly_text: str = Field(..., description="User-readable text describing the error.")
    error: str = Field(..., description="Error code")
    expected_fixes: List[str] = Field(..., description="Expected fixes to the mapping problem")
    tenant: str = Field(..., description="Tenant expected to resolve the issue.")
    goid: str | None = Field(None, description="GOID of the offending crossroads entity. Usually set if the tenant needs to fix an incoming fanout.")
    mapping_source_id: str | None = Field(None, description="The source ID of the offending mapping. Usually set if the tenant needs to fix an outgoing fanout")
    mapping_source_system: str | None = Field(None, description="The source system of the offending mapping. Only if tenant != Gravitate")
    mapping_display_name: str | None = Field(None, description="The display name of the offending mapping.")


class CrossroadsLog(BaseModel):
    """A CrossroadsLog object describes the results of a crossroads operation."""
    date: datetime = Field(description="Datetime the log was filed.")
    status: Literal["succeeded", "failed", "pending", "unknown"] = Field(
        description="Status of the logged operation. Creators should specify 'succeeded' 'failed' or 'pending'",
        default="unknown")
    tenant: str = Field(description="The tenant this record exists in.")
    record_collection: str = Field(description="The DB collection this record belongs to.")
    record_id: str = Field(description="The DB ID of the relevant record. Use with record_collection to look up data.")
    record_id_field: str = Field(description="The DB field containing the record_id.")

    target_tenant: str = Field(description="The tenant that the crossroads integration was targeting.")
    target_record_collection: str | None = Field(default=None, description="The DB collection of the record in the target tenant.")
    target_record_id: str | None = Field(default=None, description="The DB ID of the record in the target tenant.")
    target_record_id_field: str | None = Field(default=None, description="The DB field containing the record_id.")

    issue_key: Optional[str] = Field(
        description="If status='failed', this should be set with a link to a Rita Issue describing the encountered problem.",
        default=None)
    issue: Optional[IssueBase] = Field(description="If status='failed', data provided here will be used to create the issue.", default=None)
    is_mapping_failure: bool = Field(False, description="True if the reason for this log being 'failed' is a mapping error. If this is true, the mapping_error property will be set.")
    mapping_error: CrossroadsMappingError | None = Field(default=None, description="Mapping error details. Set if mapping_error == True")
    success_message: Optional[str] = Field(description="If status='success', this should be set with a message to show the user.", default=None)
    success_details: Optional[dict] = Field(
        description="If status='success', this should be set with additional details related to the successful integration. E.g. set the order number", default=None)
    worker_name: str = Field(description="The name of the worker that created this log.")
    connection_id: str | None = Field(None, description="The connection that generated this log.")
    group: str | None = Field(None, description="Freeform group field; can be used as a display grouping for logs from multiple workers. E.g. 'TTE -> Caseys Crossroads' could group logs from many workers in the UI.")
    worker_request: dict | None = Field(None, description="WIP. Plan is to use this to stash the worker request on a failure for retry capabilities.")
    correlation_id: str | None = Field(default=None, description="The correlation ID for this log.")

    def is_unspecific(self) -> bool:
        """Returns true if this log is "Unspecific" and not attached to any record."""
        return self.record_id == "unknown" and self.record_collection == "unknown" and self.record_id_field == "unknown"

    @classmethod
    def create_unspecific(cls, **kwargs):
        kwargs = kwargs | {"record_id": "unknown", "record_collection": "unknown", "record_id_field": "unknown"}
        return cls(**kwargs)


class OutputCrossroadsLog(CrossroadsLog):
    correlation_id: str | None = Field(default=None, description="The correlation ID for this crossroads log.")


class MasterReferenceAudit(BaseModel):
    """A snapshot of a MasterReferenceData or MasterReferenceLink object changed at a certain time. Creation and editing an MRD will update this object."""
    document: Any = Field(description="A copy of the document as of the time of the change.")
    doc_id: str = Field(description="MongoDB ID of the document in the relevant table.")
    user: str = Field(description="The user that updated the document.")
    date: datetime = Field(description="The datetime the document was updated.")



from enum import Enum
from typing import Optional, Dict, Union, List

from pydantic import BaseModel

from bb_integrations_lib.gravitate.model import System
from bb_integrations_lib.models.rita.mapping import Map


class Action(str, Enum):
    start = "start"
    stop = "stop"
    error = "error"


class IntegrationType(str, Enum):
    """Denotes available integration types for a Mapper"""
    sql = "sql"
    """Strictly limited to an MS SQL server database"""
    rita = "rita"
    """Strictly limited to rita mappings"""


class IntegrationMappingConfig(BaseModel):
    type: IntegrationType | System = None
    """The integration type to pull mappings from"""
    external_id_field: Optional[str] = None
    """External: to Gravitate. Denotes the id field to be used"""
    gravitate_id_field: Optional[str] = None
    """The Gravitate id field"""
    gravitate_name_field: Optional[str] = None
    external_name_field: Optional[str] = None
    query: Optional[str] = None
    """An optional query string"""


class EntityConfig(BaseModel):
    mapping_enabled: Optional[bool] = True
    mapping_integration: Optional[IntegrationMappingConfig] = {}
    external_system_integration: Optional[IntegrationMappingConfig] = {}

class PricePublisher(BaseModel):
    id: Optional[str] = None
    name: str
    price_type: str
    extend_by_days: Optional[int] = None
    contract_id_override: Optional[str] = None
    """If provided we will override the field with the value provided"""
    use_contract_id: bool = True



class PricingStrategy(str, Enum):
    use_latest = "Use Latest"
    """Only includes latest price -> may miss intraday changes"""
    use_prior_to_latest = "Use Prior to Latest"
    """Includes both the latest and latest minus one"""
    use_historic ="Use Historic"
    """Includes up to 10 historic changes for instrument since previous workday"""

    @property
    def strategy_includes(self) -> int:
        if self == PricingStrategy.use_latest:
            return 1
        if self == PricingStrategy.use_prior_to_latest:
            return 2
        if self == PricingStrategy.use_historic:
            return 10

class PricingIntegrationConfig(BaseModel):
    environment: str
    """The customer environment; i.e. TTE"""
    price_publishers: List[PricePublisher]
    """The list of price publisher from which to pull prices"""
    entity_config: Dict[str, EntityConfig] = {}
    """A key: EntityConfig pair, describing an entity config. i.e. {'products': EntityConfig}"""
    price_mapper_ttl: Optional[int] = 3600
    """The ttl cache release for the mapper"""
    price_mapper_debug_mode: Optional[bool] = False
    """Debug mode to enable verbose logging"""
    source_system: Optional[str] = None
    """The source system from where prices originate"""
    source_system_id: Optional[str] = None
    """The source system id from where prices originate"""
    strategy: Optional[PricingStrategy] = PricingStrategy.use_historic
    sql_secret_name: Optional[str] = None
    """Secret name for SQL Server credentials (used with APIFactory)"""



class Group(BaseModel):
    name: Optional[str] = None
    ids: list[str]
    length: int
    extra_data: Optional[Union[Dict, Map]] = None


class Groups(BaseModel):
    product_groups: Optional[Dict[str, Group]] = None
    location_groups: Optional[Dict[str, Group]] = None
    supplier_groups: Optional[Dict[str, Group]] = None
    price_publisher_groups: Optional[Dict[str, Group]] = None


class PriceMappings(BaseModel):
    product_mappings: Union[Dict, List]
    location_mappings: Union[Dict, List]
    supplier_mappings: Union[Dict, List]
    price_publishers: Union[Dict, List]





from typing import Optional, Dict

from pydantic import BaseModel, Field


class Config(BaseModel):
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    psk: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    base_url: Optional[str] = None
    mongo_conn_str: Optional[str] = None
    mongo_db_name: Optional[str] = None
    extra_data: Optional[Dict] = Field(default_factory=dict)


class Configs(BaseModel):
    rita: Optional[Config] = Field(default_factory=Config)
    sd: Optional[Config] = Field(default_factory=Config)
    pe: Optional[Config] = Field(default_factory=Config)
    crossroads: Optional[Config] = Field(default_factory=Config)


class GlobalConfig(BaseModel):
    prod: Configs = Field(default_factory=Configs)
    test: Configs = Field(default_factory=Configs)
    extra_data: Optional[Dict] = Field(default_factory=dict)



import json
from typing import List

from loguru import logger

from bb_integrations_lib.gravitate.pe_api import GravitatePEAPI
from bb_integrations_lib.models.pipeline_structs import BBDUploadResult, UploadResult
from bb_integrations_lib.models.rita.issue import IssueBase, IssueCategory
from bb_integrations_lib.protocols.flat_file import PePriceMergeIntegration
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.util.utils import CustomJSONEncoder


class PEPriceMerge(Step):
    def __init__(
            self,
            pe_client: GravitatePEAPI,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.pe_client = pe_client

    def describe(self) -> str:
        return "Merge Prices in Pricing Engine"

    async def execute(self, i: List[PePriceMergeIntegration]) -> BBDUploadResult:
        failed_rows: List = []
        success_rows: List = []
        responses: List = []
        try:
            for row in i:
                row_dump = row.model_dump(exclude_none=True)
                try:
                    response = await self.pe_client.merge_prices(row_dump)
                    response_data = response.json()
                    success_rows.append({**row_dump, "response": response_data})
                    responses.append(response_data)
                except Exception as e:
                    logger.error(f"Failed to merge row: {e}")
                    failed_rows.append(row_dump)
                    continue
        except Exception as e:
            logger.error(f"Failed to merge rows: {e}")
            raise e

        logs = {
            "request": [l.model_dump() for l in i],
            "response": responses
        }
        self.pipeline_context.included_files["merge prices logs"] = json.dumps(logs, cls=CustomJSONEncoder)
        return UploadResult(succeeded=len(success_rows), failed=len(failed_rows),
                            succeeded_items=list(success_rows))




from datetime import datetime
from enum import Enum
from typing import Optional, List, Annotated

from annotated_types import MaxLen
from pydantic import BaseModel, Field


class IssueCategory(Enum):
    UNKNOWN = "unknown"
    MISC = "misc"
    ORDER = "order"
    PAYROLL = "payroll"
    TANK_READING = "tank"
    PRICE = "price"
    REFERENCE_DATA = "reference_data"
    CROSSROADS = "crossroads"
    CROSSROADS_MAPPINGS = "crossroads_mappings"


class IssueBase(BaseModel):
    key: str = Field(description="Unique key for this issue")
    config_id: str = Field(description="Config object this issue is reported under.")
    # group or config id?
    name: str = Field(description="Human readable name")
    category: IssueCategory = Field(description="Broad category of the issue")
    problem_short: Optional[str] = Field(default=None,
                                         description="Short description of the problem (e.g. an exception class name)")
    problem_long: Optional[str] = Field(default=None, description="Long description of the problem (e.g. traceback)")
    occurrences: Annotated[
        List[datetime], Field(default=[], description="List of most recent datetimes that this issue was reported",
                              max_length=50)]

    class Config:
        arbitrary_types_allowed = True

    @property
    def most_recent_occurrence(self) -> Optional[datetime]:
        return max(self.occurrences, default=None)

    def get_occurrence_limit(self) -> int:
        """Get the annotated max_length of the occurrences list."""
        for meta in self.model_fields["occurrences"].metadata:
            if isinstance(meta, MaxLen):
                return meta.max_length
        raise AttributeError("IssueBase.occurrences does not have an annotated max_length")

    def trim_occurrences(self):
        """
        Trims the occurrences array to fit within the annotated Field.max_length.
        The array gets sorted as a side effect.
        """
        self.occurrences = sorted(self.occurrences)[-self.get_occurrence_limit():]


class UpdateIssue(BaseModel):
    key: str = Field(description="Unique key for this issue")
    config_id: Optional[str] = Field(default=None, description="Config object this issue is reported under.")
    name: Optional[str] = Field(default=None, description="Human readable name")
    category: Optional[IssueCategory] = Field(default=None, description="Broad category of the issue")
    problem_short: Optional[str] = Field(default=None,
                                         description="Short description of the problem (e.g. an exception class name)")
    problem_long: Optional[str] = Field(default=None, description="Long description of the problem (e.g. traceback)")



from typing import List, Union

from bb_integrations_lib.models.pipeline_structs import BolExportResults
from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.provider.ftp.client import FTPIntegrationClient
from bb_integrations_lib.shared.model import FileReference, RawData
from .sftp_export_file_step import SFTPExportFileStep


class SFTPExportManyFilesStep(Step):
    def __init__(
            self,
            ftp_client: FTPIntegrationClient,
            ftp_destination_dir: str,
            field_sep: str = ",",
            allow_empty: bool = False,
            *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.sftp_export_step = SFTPExportFileStep(
            ftp_client=ftp_client,
            ftp_destination_dir=ftp_destination_dir,
            field_sep=field_sep,
            allow_empty=allow_empty,
        )

    def describe(self) -> str:
        return "SFTP Many Files Export"

    async def execute(self, files: List[Union[FileReference, RawData, BolExportResults]]) -> List[FileReference]:
        results = []
        for file in files:
            result = await self.sftp_export_step.execute(file)
            if result:
                results.append(result)
        return results


from datetime import datetime
from typing import AsyncGenerator

import pandas as pd
from loguru import logger

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.mappers.rita_mapper import RitaMapperCore, RitaMapper, RitaAPIMappingProvider
from bb_integrations_lib.models.rita.mapping import MappingType
from bb_integrations_lib.protocols.pipelines import GeneratorStep
from bb_integrations_lib.provider.api.cargas.model import CreateWholesaleTicketRequestBundle, \
    CreateWholesaleTicketRequest, CreateWholesaleLineRequest


class ConvertBBDOrderToCargasWholesaleStep(GeneratorStep):
    def __init__(self, rita_client: GravitateRitaAPI, sd_client: GravitateSDAPI, mapping_source_system: str = "Cargas", *args, **kwargs):
        """
        Convert a BBD order's drops to Cargas wholesale ticket request objects. Input should be an item returned by the
        v1/bols_and_drops endpoint, and should only include orders with at least 1 drop at a wholesale store. Passing an
        order with no wholesale drops will result in a ticket request being generated with no line items.

        Mappings are best-effort; if a field that is not strictly required is not mapped, it will be left out.

        :param rita_client: The GravitateRitaAPI instance to use to get mappings with.
        :param sd_client: The GravitateSDAPI instance to use for API calls.
        :param mapping_source_system: The source system to use for mapping. Defaults to "Cargas".
        """
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.sd_client = sd_client
        self.mapping_source_system = mapping_source_system

        self.store_lkp = {}
        self.driver_sched_lkp = {}
        # Only optional because it will be late initialized by either _load_lookups or _set_test_lookups.
        self.rita_mapper: RitaMapper | None = None

    def describe(self) -> str:
        return f"Convert drops on a BBD order to Cargas create wholesale ticket / ticket line requests"

    async def _load_lookups(self, driver_schedules: set[str]):
        stores = await self.sd_client.get_all_stores()
        stores.raise_for_status()
        self.store_lkp = {store["_id"]: store for store in stores.json()}

        drivers = await self.sd_client.get_driver_tracking(driver_schedule_ids=list(driver_schedules))
        drivers.raise_for_status()
        self.driver_sched_lkp = {driver["driver_schedule_id"]: driver for driver in drivers.json()}

        self.rita_mapper = RitaMapper(
            provider=RitaAPIMappingProvider(self.rita_client),
            source_system=self.mapping_source_system
        )
        await self.rita_mapper.load_mappings_async()

    def _set_test_lookups(self, store_lkp: dict, driver_sched_lkp: dict, rita_mapper: RitaMapperCore):
        self.store_lkp = store_lkp
        self.driver_sched_lkp = driver_sched_lkp
        self.rita_mapper = rita_mapper

    async def generator(self, orders: list[dict]) -> AsyncGenerator[CreateWholesaleTicketRequestBundle, None]:
        driver_schedules = set([
            abol["driver_schedule"] for order in orders for abol in order["allocated_bols"]
        ])
        await self._load_lookups(driver_schedules=driver_schedules)

        assert self.rita_mapper is not None, "RITA mapper must be initialized before generator is called"

        for order in orders:
            for result in self.convert_order(order):
                yield result

    def convert_order(self, order: dict) -> list[CreateWholesaleTicketRequestBundle]:
        """
        Converts an S&D order dictionary into one or more Cargas ticket request bundle objects, including line items.
        Line items produced by this function will have a blank DocumentID, which should be filled in with the
        DocumentID that CreateWholesaleTicket returns.
        """
        order_number = order["order_number"]
        allocated_bols = order["allocated_bols"]
        if not allocated_bols:
            raise Exception("Could not convert order: no allocated BOLs")

        bols_df = pd.DataFrame.from_records(allocated_bols)
        # Precompute destination counterparty - could be 2 different stores of same counterparty on a split load
        bols_df["store_counterparty_name"] = bols_df.apply(
            lambda x: self.store_lkp[x["store_id"]]["counterparty_name"],
            axis="columns"
        )
        bols_df["driver_name"] = bols_df.apply(
            lambda x: self.driver_sched_lkp.get(x["driver_schedule"], {}).get("driver_name", ""),
            axis="columns"
        )
        bols_df["tractor"] = bols_df.apply(
            lambda x: self.driver_sched_lkp.get(x["driver_schedule"], {}).get("driver_log", {}).get("tractor", ""),
            axis="columns"
        )

        results = []
        bol_gb = bols_df.groupby(by=["bol_number", "store_counterparty_name", "driver_name", "tractor"])
        for (bol_number, store_cp_name, driver_name, tractor), group in bol_gb:
            mcid_str = None
            try:
                mcid_str = self.rita_mapper.get_source_parent_id(store_cp_name, MappingType.counterparty)
                mapped_customer_id = int(mcid_str)
            except KeyError as e:
                mapped_customer_id = None
                logger.warning(f"Failed to map store counterparty {store_cp_name}: {e}, skipping drops")
                continue
            except ValueError as e:
                logger.warning(f"Failed to convert mapped customer ID '{mcid_str}' to int, skipping drops")
                continue

            try:
                mapped_driver_id = self.rita_mapper.get_source_parent_id(driver_name, MappingType.driver)
            except KeyError:
                logger.warning(f"Failed to map driver ID '{driver_name}', it will be blank on the ticket")
                mapped_driver_id = None
            try:
                mapped_tractor_id = self.rita_mapper.get_source_parent_id(tractor, MappingType.tractor)
            except KeyError:
                logger.warning(f"Failed to map tractor ID '{tractor}', it will be blank on the ticket")
                mapped_tractor_id = None

            try:
                ticket_request = CreateWholesaleTicketRequest(
                    CustomerID=mapped_customer_id,
                    DeliveryDate=datetime.fromisoformat(max(group["delivered_date"])),
                    CustomerPONumber=bol_number,
                    DriverID=mapped_driver_id,
                    WholesaleTruckID=mapped_tractor_id,
                    Message="",
                    InvoiceNotes="",
                    CostCenterID=None,
                    SubTypeID=None,
                    SalespersonID=None,
                    AdditionalNotes="",
                    UserName="Gravitate"
                )

                line_requests = []
                for index, bol_item in group.iterrows():
                    try:
                        store_number = bol_item["store_number"]
                        tank_number = str(bol_item["store_tank"])
                        mtid_str = None
                        try:
                            mtid_str = self.rita_mapper.get_source_child_id(
                                gravitate_parent_id=store_number, gravitate_child_id=tank_number,
                                mapping_type=MappingType.tank
                            )
                            mapped_tank_id = int(mtid_str)
                        except KeyError as e:
                            logger.warning(f"Failed to map tank '{tank_number}', skipping BOL item: {e}")
                            continue
                        except ValueError as e:
                            logger.warning(
                                f"Failed to convert mapped tank '{tank_number}'->'{mtid_str}' to int, skipping BOL item: {e}")
                            continue

                        miid_str = None
                        bol_product = bol_item["bol_product"]
                        try:
                            miid_str = self.rita_mapper.get_source_parent_id(
                                gravitate_id=bol_product, mapping_type=MappingType.product
                            )
                            mapped_item_id = int(miid_str)
                        except KeyError as e:
                            logger.warning(f"Failed to map product '{bol_product}', skipping BOL item: {e}")
                            continue
                        except ValueError as e:
                            logger.warning(
                                f"Failed to convert mapped product '{bol_product}'->'{miid_str}', skipping BOL item: {e}")
                            continue

                        # We could theoretically be handed an order with no wholesale store drops, in which case we'll
                        # upload a wholesale ticket with no line items. However, upstream should prefilter out any
                        # orders with no wholesale store drops.
                        if self.store_lkp[bol_item["store_id"]].get("extra_data", {}).get("type") != "Wholesale":
                            logger.info(
                                f"Skipping drop at non-wholesale store '{group['store_number']}' "
                                f"(order {order_number}, BOL {bol_number})"
                            )
                            continue

                        line_requests.append(
                            CreateWholesaleLineRequest(
                                DocumentID=-1,
                                ItemID=mapped_item_id,
                                TankID=mapped_tank_id,
                                Quantity=bol_item["bol_net_volume_allocated"],
                                QuantityGross=bol_item["bol_gross_volume_allocated"],
                                UnitPrice=bol_item.get("price", {}).get("price"),
                                FreightRateID=None,  # TODO: Retrieve? May not be necessary
                                VendorLocationID=None,
                                FreightAmount=None,  # TODO: Retrieve? May not be necessary
                                SurchargeAmount=None,
                                UnitCostOverride=None,
                                CustomerPricingID=None,
                                UserName="Gravitate"
                            )
                        )
                    except Exception as e:
                        logger.warning(f"Could not convert bol item #{index} on order {order_number}: {e}")
                        continue

                if not line_requests:
                    logger.warning(
                        f"No line requests for order {order_number}; will upload a ticket request but it will be empty")
                results.append(CreateWholesaleTicketRequestBundle(
                    ticket_request=ticket_request,
                    line_requests=line_requests
                ))

            except Exception as e:
                logger.warning(f"Could not convert order {order_number}, BOL {bol_number}: {e}")
                continue

        return results

    def get_store_cp_name(self, store_id: str) -> str:
        return self.store_lkp[store_id]["counterparty_name"]



from enum import Enum
from functools import lru_cache
from pprint import pprint
from types import NoneType
from typing import List, Dict, Literal, Union, override, get_args

from pydantic import BaseModel, Field

from bb_integrations_lib.models.rita.crossroads_entities import CrossroadsCompany, CrossroadsProduct, \
    CrossroadsTerminal, CrossroadsSite, CrossroadsTank, BaseCrossroadsEntity
from bb_integrations_lib.models.rita.crossroads_monitoring import CrossroadsMappingError
from bb_integrations_lib.shared.model import AgGridBaseModel


class CrossroadsMappingType(str, Enum):
    """These are the types of records we support linking. """
    location = "location"
    product = "product"
    site = "site"
    tank = "tank"
    counterparty = "counterparty"
    terminal = "terminal"

    def get_sd_collection(self):
        if self in _sd_collections:
            return _sd_collections[self]
        else:
            return None

    @classmethod
    def get_parent_types(self):
        return _sd_collections.keys()

    def get_sd_id_field(self):
        if self not in _sd_id_fields:
            return "id"
        return _sd_id_fields[self]

    def get_sd_name_field(self):
        if self not in _sd_name_fields:
            return "name"
        return _sd_name_fields[self]

_sd_collections = {
    CrossroadsMappingType.location: "location",
    CrossroadsMappingType.product: "product",
    CrossroadsMappingType.site: "store",
    CrossroadsMappingType.counterparty: "counterparty",
    CrossroadsMappingType.terminal: "location",
}

_sd_name_fields = {
    CrossroadsMappingType.tank: "tank_id",
}

_sd_id_fields = {
    CrossroadsMappingType.site: "store_number"
}


class CoreCrossroadsMapping(AgGridBaseModel):
    """
    An instance of this object associates a record in a source system to zero or more GOIDs. The data here lives only in
    client tenants, not in the master tenant database. A CrossroadsMapping may be incomplete (matches no GOIDs), exact
    (matches exactly 1 GOID), or multiple (matches 2+ GOIDs)
    """
    type: CrossroadsMappingType = Field(..., description="The type of record.")
    goids: List[str] = Field([], description="Linked CrossroadsEntities by their GOIDs. A mapping may link to multiple CrossroadsEntities.")
    display_name: str = Field(..., description="This is the friendly name of this record. It will by synced with the datasource if available.")
    source_id: str = Field(..., description="Either (1) The Mongodb ID of the item in the tenant's S&D instance, or (2) The ID of the item in the specified source system.")
    matching_info: dict = Field({}, description="This is info that we expect to use an AI agent to help us match records with CrossroadsEntities.")
    extra_data: dict = Field({}, description="Additional data that may be used by mapping logic as needed.")



class CrossroadsMapping(CoreCrossroadsMapping):
    source_system: str = Field("Gravitate Supply & Dispatch", description="The name of the system that this Mapping is for. Defaults to Gravitate Supply & Dispatch")
    children: Dict[str, CoreCrossroadsMapping] = Field({}, description="Child mappings keyed by their source_id")


class CrossroadsMappingResult(AgGridBaseModel):
    origin_tenant: str = ""
    origin_source_id: str = ""
    origin_source_system: str = ""
    origin_crossroadsmapping: CrossroadsMapping = {}
    target_tenant: str = ""
    target_crossroadsmappings: list[CrossroadsMapping] = []
    matched_crossroadsentities: dict[str, dict] = {}
    milliseconds_taken: int | None = None


class MappingRequestData(BaseModel):
    """
    The body of a mapping request. A discriminated union. There is an is_resolved field that the backend uses to know if it can look in the non-id fields for more data.

    AUTOMAGICAL: When defining a sublcass, any fields suffixed with _id will be auto-resolved by the Backend mapping methods.
    The resolved data will be put into a field of the same name without the _id. E.g. `supplier_id` and `supplier` form a
    pair that will be auto matched.
    """
    type: str
    is_resolved: bool | None = Field(default=None, description="Set 'true' by the mapping engine if all of the id fields have been resolved. This is computed by backend; user input has no effect here.")

    @classmethod
    def automap_fields(cls) -> dict[str, str]:
        """Returns a map with _id fields as the keys and full fields as the values. For use with getattr"""
        id_fields = [f for f in cls.model_fields.keys() if f.endswith("_id")]
        result = {}
        for id_field_name in id_fields:
            full_field_name = id_field_name.rstrip("_id")
            if full_field_name in cls.model_fields:
                result[id_field_name] = full_field_name
        return result

    @classmethod
    def automap_child_fields(cls) -> dict[str, str]:
        """Returns a map with _cid fields as the keys and full fields as the values. For use with getattr"""
        cid_fields = [f for f in cls.model_fields.keys() if f.endswith("_cid")]
        result = {}
        for id_field_name in cid_fields:
            full_field_name = id_field_name.rstrip("_cid")
            if full_field_name in cls.model_fields:
                result[id_field_name] = full_field_name
        return result

    @classmethod
    def get_crossroads_entity_type(cls, full_field_name: str):
        field = cls.model_fields[full_field_name]
        for arg in get_args(field.annotation):
            if arg == NoneType or issubclass(arg, CoreCrossroadsMapping):
                continue
            return arg

    def resolve_child_mappings(self, tenant: str) -> None:
        """Called after all parent mappings have been resolved. If there are children mappings, override this method and
        implement a method to fetch them from the parent mappings. This method can
        :param tenant: When an error is thrown, the provided tenant will be linked to the error.
        """
        pass

    def get_parent_mapping(self, field_name: str) -> CrossroadsMapping:
        """May be called by the mapping engine after all parent mappings are resolved. This method can be provided either
        the ID field name or the full field name of a child mapping and should return the contents of the parent mapping object."""
        pass

    def get_mapping_type_for_field(self, field_name) -> CrossroadsMappingType | None:
        """
        Called when attempting to auto-resolve multiple mappings that have different types. This is a common case, so
        we attempt to determine which data type we should use. E.g. if a crossroads entity is linked to 2 mappings, 1 site
        and 1 location, this method should provide the way to disambiguate which one this property expects based on the field.
        If the logic is more complicated than "always use this type for this mapping request" then this method is not the right
        spot to implement that info; use custom rules instead.
        """
        return None

    def access_by_property_name(self, property_name: str) -> str:
        try:
            path = property_name.split(".")
            obj = self
            for p in path:
                obj = getattr(obj, p)
            return str(obj)
        except Exception as e:
            return "<<<UNKNOWN>>>"

    @classmethod
    @lru_cache(maxsize=16)
    def outgoing_rule_properties(cls) -> list[str]:
        """
        Returns the valid rule properties for this request data when considering outgoing rules. These properties can be used in conditionals.
        Because they're restricted to outgoing rules you should return properties related to the crossroads mappings.
        """
        fields = list(cls.automap_fields().items()) + list(cls.automap_child_fields().items())
        results = []
        for _, full_field_name in fields:
            results += [f"{full_field_name}.source_id", f"{full_field_name}.display_name"]
        return results


    @classmethod
    @lru_cache(maxsize=16)
    def incoming_rule_properties(cls) -> list[str]:
        """
        Returns the valid rule properties for this request data when considering outgoing rules. These properties can be used in conditionals.
        Because they're restricted to outgoing rules you should return properties related to the crossroads mappings.
        """
        fields = list(cls.automap_fields().items()) + list(cls.automap_child_fields().items())
        results = []
        for _, full_field_name in fields:
            pydantic_field = cls.model_fields[full_field_name]
            for arg in get_args(pydantic_field.annotation):
                if arg == NoneType or issubclass(arg, CoreCrossroadsMapping):
                    continue
                props = list(arg.model_fields.keys())
                props = [p for p in props if p != "is_active" and p != "record_owner"]
                results += [f"{full_field_name}.{p}" for p in props]
        return results


class BasicMapping(MappingRequestData):
    """A basic mapping from one tenant to another. Rules are very limited here since there's very little data."""
    type: Literal["basic_mapping"] = "basic_mapping"
    obj_id: str
    obj: Union[CrossroadsMapping, BaseCrossroadsEntity, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by obj_id")
    req_type: CrossroadsMappingType | None = None

    @override
    def get_mapping_type_for_field(self, field_name) -> CrossroadsMappingType | None:
        if self.req_type is None:
            return None
        return self.req_type


class LoadPlan(MappingRequestData):
    type: Literal["load_plan"] = "load_plan"
    supplier_id: str
    product_id: str
    terminal_id: str
    destination_id: str
    destination_tank_cid: str
    supplier: Union[CrossroadsMapping, CrossroadsCompany, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by supplier_id")
    product: Union[CrossroadsMapping, CrossroadsProduct, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by product_id")
    terminal: Union[CrossroadsMapping, CrossroadsTerminal, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by terminal_id")
    destination: Union[CrossroadsMapping, CrossroadsSite, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by destination_id")
    destination_tank: Union[CoreCrossroadsMapping, CrossroadsTank, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at tank_id")


    @override
    def get_mapping_type_for_field(self, field_name) -> CrossroadsMappingType | None:
        if field_name.startswith("product"):
            return CrossroadsMappingType.product
        elif field_name.startswith("terminal"):
            return CrossroadsMappingType.terminal
        elif field_name.startswith("supplier"):
            return CrossroadsMappingType.counterparty
        elif field_name.startswith("destination_tank"):
            return CrossroadsMappingType.tank
        elif field_name.startswith("destination"):
            return CrossroadsMappingType.site
        raise ValueError(f"Invalid argument: {field_name}")

    @override
    def resolve_child_mappings(self, tenant: str) -> None:
        if self.destination_tank_cid.startswith("tank:"):
            tanks = [self.destination.children[t] for t in self.destination.children if self.destination_tank_cid in self.destination.children[t].goids and self.destination.children[t].type == CrossroadsMappingType.tank]
            if len(tanks) != 1:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for destination_tank_cid = {self.destination_tank_cid}. Could not find a unique child of site {self.destination_id} that had {self.destination_tank_cid} as its GOID. Update the child mapping on the site to link to the Crossroads entity with this GOID.",
                    error="no link to child mapping", expected_fixes=["update child mapping"], tenant=tenant,
                )
            self.destination_tank_cid = tanks[0].source_id
            self.destination_tank = tanks[0]
        else:
            self.destination_tank = self.destination.children.get(self.destination_tank_cid)
            if not self.destination_tank:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for destination_tank_cid = {self.destination_tank_cid}. There was no child with source_id {self.destination_tank_cid} for the site {self.destination_id}. Create the child mapping on the site and link it to Crossroads.",
                    error="no child mapping", expected_fixes=["create child mapping"], tenant=tenant,
                )

    @override
    def get_parent_mapping(self, field_name: str) -> CrossroadsMapping:
        if field_name.startswith("destination_tank"):
            return self.destination
        raise ValueError("Invalid argument")


class SpecificSupplyDrop(MappingRequestData):
    type: Literal["drop"] = "drop"
    product_id: str
    destination_id: str
    destination_tank_cid: str
    product: Union[CrossroadsMapping, CrossroadsProduct, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by product_id")
    destination: Union[CrossroadsMapping, CrossroadsSite, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by destination_id")
    destination_tank: Union[CoreCrossroadsMapping, CrossroadsTank, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at tank_id")

    @override
    def resolve_child_mappings(self, tenant: str) -> None:
        if self.destination_tank_cid.startswith("tank:"):
            tanks = [self.destination.children[t] for t in self.destination.children if self.destination_tank_cid in self.destination.children[t].goids and self.destination.children[t].type == CrossroadsMappingType.tank]
            if len(tanks) != 1:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for destination_tank_cid = {self.destination_tank_cid}. Could not find a unique child of site {self.destination_id} that had {self.destination_tank_cid} as its GOID. Update the child mapping on the site to link to the Crossroads entity with this GOID.",
                    error="no link to child mapping", expected_fixes=["update child mapping"], tenant=tenant,
                )
            self.destination_tank_cid = tanks[0].source_id
            self.destination_tank = tanks[0]
        else:
            self.destination_tank = self.destination.children.get(self.destination_tank_cid)
            if not self.destination_tank:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for destination_tank_cid = {self.destination_tank_cid}. There was no child with source_id {self.destination_tank_cid} for the site {self.destination_id}. Create the child mapping on the site and link it to Crossroads.",
                    error="no child mapping", expected_fixes=["create child mapping"], tenant=tenant,
                )

    @override
    def get_parent_mapping(self, field_name: str) -> CrossroadsMapping:
        if field_name.startswith("destination_tank"):
            return self.destination
        raise ValueError("Invalid argument")

    @override
    def get_mapping_type_for_field(self, field_name) -> CrossroadsMappingType | None:
        if field_name.startswith("product"):
            return CrossroadsMappingType.product
        elif field_name.startswith("supplier"):
            return CrossroadsMappingType.counterparty
        elif field_name.startswith("destination_tank"):
            return CrossroadsMappingType.tank
        elif field_name.startswith("destination"):
            return CrossroadsMappingType.site
        raise ValueError(f"Invalid argument: {field_name}")


class TSDDrop(MappingRequestData):
    type: Literal["tsd_drop"] = "tsd_drop"
    destination_id: str
    destination_tank_cid: str
    destination: Union[CrossroadsMapping, CrossroadsSite, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by destination_id")
    destination_tank: Union[CoreCrossroadsMapping, CrossroadsTank, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at tank_id")

    @override
    def resolve_child_mappings(self, tenant: str) -> None:
        if self.destination_tank_cid.startswith("tank:"):
            tanks = [self.destination.children[t] for t in self.destination.children if
                     self.destination_tank_cid in self.destination.children[t].goids and self.destination.children[t].type == CrossroadsMappingType.tank]
            if len(tanks) != 1:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for destination_tank_cid = {self.destination_tank_cid}. Could not find a unique child of site {self.destination_id} that had {self.destination_tank_cid} as its GOID. Update the child mapping on the site to link to the Crossroads entity with this GOID.",
                    error="no link to child mapping", expected_fixes=["update child mapping"], tenant=tenant,
                )
            self.destination_tank_cid = tanks[0].source_id
            self.destination_tank = tanks[0]
        else:
            self.destination_tank = self.destination.children.get(self.destination_tank_cid)
            if not self.destination_tank:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for destination_tank_cid = {self.destination_tank_cid}. There was no child with source_id {self.destination_tank_cid} for the site {self.destination_id}. Create the child mapping on the site and link it to Crossroads.",
                    error="no child mapping", expected_fixes=["create child mapping"], tenant=tenant,
                )

    @override
    def get_parent_mapping(self, field_name: str) -> CrossroadsMapping:
        if field_name.startswith("destination_tank"):
            return self.destination
        raise ValueError("Invalid argument")

    @override
    def get_mapping_type_for_field(self, field_name) -> CrossroadsMappingType | None:
        if field_name.startswith("destination_tank"):
            return CrossroadsMappingType.tank
        elif field_name.startswith("destination"):
            return CrossroadsMappingType.site
        raise ValueError(f"Invalid argument: {field_name}")


class TankReading(MappingRequestData):
    type: Literal["tank_reading"] = "tank_reading"
    site_id: str
    tank_cid: str
    site: Union[CrossroadsMapping, CrossroadsSite, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at site_id")
    tank: Union[CoreCrossroadsMapping, CrossroadsTank, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at tank_id")

    @override
    def resolve_child_mappings(self, tenant: str) -> None:
        if self.tank_cid.startswith("tank:"):
            tanks = [self.site.children[t] for t in self.site.children if self.tank_cid in self.site.children[t].goids and self.site.children[t].type == CrossroadsMappingType.tank]
            if len(tanks) != 1:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for tank_cid = {self.tank_cid}. Could not find a unique child of site {self.site_id} that had {self.tank_cid} as its GOID. Update the child mapping on the site to link to the Crossroads entity with this GOID.",
                    error="no link to child mapping", expected_fixes=["update child mapping"], tenant=tenant,
                )
            self.tank_cid = tanks[0].source_id
            self.tank = tanks[0]
        else:
            self.tank = self.site.children.get(self.tank_cid)
            if not self.tank:
                raise CrossroadsMappingError(
                    friendly_text=f"Error while resolving child mappings for tank_cid = {self.tank_cid}. There was no child with source_id {self.tank_cid} for the site {self.site_id}. Create the child mapping on the site and link it to Crossroads.",
                    error="no child mapping", expected_fixes=["create child mapping"], tenant=tenant,
                )

    @override
    def get_parent_mapping(self, field_name: str) -> CrossroadsMapping:
        if field_name.startswith("tank"):
            return self.site
        raise ValueError("Invalid argument")

    @override
    def get_mapping_type_for_field(self, field_name: str) -> CrossroadsMappingType | None:
        if field_name.startswith("site"):
            return CrossroadsMappingType.site
        elif field_name.startswith("tank"):
            return CrossroadsMappingType.tank
        raise ValueError(f"Invalid argument: {field_name}")


class OrderBasics(MappingRequestData):
    type: Literal["order_basics"] = "order_basics"
    supplier_id: str
    supplier: Union[CrossroadsMapping, CrossroadsCompany, None] = Field(default=None, description="If is_resolved == True, this will be the item pointed at by supplier_id")

    @override
    def get_mapping_type_for_field(self, field_name: str) -> CrossroadsMappingType | None:
        if field_name.startswith("supplier"):
            return CrossroadsMappingType.counterparty
        raise ValueError(f"Invalid argument: {field_name}")


class MappingRequest(BaseModel):
    """These are 'fat' mapping requests: they request backend to map multiple fields but serve as their own context for
    the rules engine."""
    origin_tenant: str
    origin_source_system: str
    target_tenant: str
    target_source_system: str

    data: Union[
        BasicMapping, LoadPlan, SpecificSupplyDrop, TSDDrop, TankReading, OrderBasics
    ] = Field(..., discriminator="type", description="The body of the mapping request. See MappingRequestData and its subclasses.")

    extra_data: dict = Field(default={}, description="Additional data that may be used by the mapping engine. Rules may also add fields here and read fields from here.")


class MappingRequestInternal(MappingRequest):
    """Holding type to differentiate from a mapping request into or out of a tenant. Has a field to mark the difference"""
    is_crossroads: Literal[True] = True





# Definitions and tech for the mapping rules engine.
from enum import Enum
from typing import Literal, runtime_checkable, Protocol

from bb_integrations_lib.models.rita.crossroads_mapping import MappingRequest
from pydantic import BaseModel, Field, model_validator, computed_field

from bb_integrations_lib.shared.model import AgGridBaseModel


class Condition(BaseModel):
    """A single condition for a property or value in a mapping rule. A rule may have many conditions."""
    property: str = Field(..., description="The property that this condition will inspect")
    predicate: Literal["equals"] = Field("equals", description="The method of comparing the property and the value.")
    value: str = Field(..., description="The value that this condition will inspect the property for, using the predicate.")


class Action(BaseModel):
    type: Literal["choose result", "inject metadata"] = Field("choose result", description="Type of action")
    value: str = Field(..., description="The value related to the type of action. This may be a GOID, a mapping ID, or a json string of metadata to provide to further mappings.")


class CrossroadsRule(AgGridBaseModel):
    """Represents rules that can describe how to resolve mappings with more than one possible resolution. Rules are either
    "incoming", "outgoing", or "override". Incoming rules describe how to choose a CrossroadsMapping when given a CrossroadsEntity.
    Outgoing rules describe how to choose a CrossroadsEntity when given a CrossroadsMapping. Override rules describe how
    to choose a CrossroadsMapping given a CrossroadsMapping, bypassing the CrossroadsEntities entirely.
    """
    display_name: str = Field("")
    type: Literal["incoming", "outgoing", "override"] = Field("incoming", description="The type of this rule. Used to choose when a rule might apply.")
    mapping_id: str | None = Field(None, description="If this rule is outgoing and related to a single mapping, then this field will be set with the DB ID of that mapping.")
    entity_goid: str | None = Field(None, description="If this rule is incoming and related to a single master entity, then this field will be set with that entity's GOID.")
    tenants: list[str] | None = Field(None, description="Which tenants this rule may apply to. If empty, can be any tenant. Must be set if type == override")
    conditions: list[Condition] = Field([], description="The match conditions that make up this rule. Logical AND: A candidate must pass all conditions.")
    action: Action = Field(..., description="The action taken by this rule when the match conditions are met.")
    conditions_mode: Literal["and", "or"] = Field(default="and", description="How should multiple conditions be applied? In AND mode the conditions will be ANDed together logically. In OR mode the conditions will be ORed together logically.")
    is_active: bool = Field(True)

    @computed_field
    @property
    def requires_properties(self) -> list[str]:
        return list({c.property for c in self.conditions})

    def can_apply(self, req: MappingRequest):
        if self.tenants is not None and len(self.tenants) > 0 and req.target_tenant not in self.tenants:
            return False
        for condition in self.conditions:
            if self.type == "outgoing" and condition.property not in req.data.outgoing_rule_properties():
                return False
            elif self.type == "incoming" and condition.property not in req.data.incoming_rule_properties():
                return False
        return True

    def meets_conditions(self, req: MappingRequest):
        if self.conditions_mode == "and":
            for condition in self.conditions:
                request_value = req.data.access_by_property_name(condition.property)
                if condition.predicate == "equals" and request_value != condition.value:
                    return False
            return True  # All conditions passed
        else:
            for conditions in self.conditions:
                request_value = req.data.access_by_property_name(conditions.property)
                if conditions.predicate == "equals" and request_value == conditions.value:
                    return True
            return False # None of the conditions were met



    @model_validator(mode="after")
    def validate(self):
        if self.mapping_id is not None and self.type != "outgoing":
            raise ValueError("`mapping_id` should only be set when type is outgoing")
        if self.entity_goid is not None and self.type != "incoming":
            raise ValueError("`entity_goid` should only be set when type is incoming")
        if not self.tenants and self.type == "override":
            raise ValueError("`tenants` must be set when type is override")
        return self





import os
from typing import Dict, Any

import loguru

from bb_integrations_lib.protocols.pipelines import Step
from bb_integrations_lib.provider.ftp.client import FTPIntegrationClient
from bb_integrations_lib.shared.model import FileConfigRawData, RawData


class DeleteSFTPStep(Step):
    def __init__(self, ftp_client: FTPIntegrationClient, src_directory: str | None = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.ftp_client = ftp_client
        self.src_directory = src_directory

    def describe(self) -> str:
        return "SFTP Delete Step"

    async def execute(self, i: Any) -> Any:
        if isinstance(i, FileConfigRawData):
            filename = os.path.join(i.file_config.inbound_directory, i.file_name)
        elif isinstance(i, RawData):
            if self.src_directory is None:
                raise RuntimeError("Attempted to delete a RawData object but src_directory was not set.")
            filename = os.path.join(self.src_directory, i.file_name)
        else:
            raise NotImplementedError(f"Unsupported input type: {type(i)}")
        try:
            self.ftp_client.delete_file(filename)
        except Exception as e:
            loguru.logger.warning(f"Failed to delete: {e}")
        return i



from typing import Protocol, Optional, Dict, Any
from pymongo.database import Database


class System(Protocol):
    sd: str
    pe: str
    rita: str


class GravitateConfig(Protocol):
    system_psk: Optional[str]
    url: Optional[str]
    conn_str: Optional[str]
    dbs: Optional[Dict[str, str]]
    system: Optional[System]
    short_name: Optional[str]
    admin_username: Optional[str]
    admin_password: Optional[str]


class SecureWebAPI(Protocol):
    system_name: str
    username: Optional[str]
    password: Optional[str]
    client_id: Optional[str]
    client_secret: Optional[str]
    psk: str
    base_url: str
    token: Optional[str]
    system: Any
    extra_headers: Dict[str, str]

    def post(self, url: str, params: Optional[Dict] = None, headers: Optional[Dict] = None,
             data: Optional[Dict] = None, json: Optional[Dict] = None, files: Optional[Dict] = None,
             timeout: int = 180, raise_error: bool = True) -> Any: ...

    def get(self, url: str, params: Optional[Dict] = None, headers: Optional[Dict] = None,
            timeout: int = 180, raise_error: bool = True) -> Any: ...

    def token_post(self, url: str, params: Optional[Dict] = None, data: Optional[Dict] = None,
                   json: Optional[Dict] = None, files: Optional[Dict] = None, timeout: int = 180,
                   raise_error: bool = True, headers: Optional[Dict] = None) -> Any: ...

    def token_get(self, url: str, params: Optional[Dict] = None, timeout: int = 180,
                  raise_error: bool = True) -> Any: ...

    def psk_post(self, url: str, params: Optional[Dict] = None, data: Optional[Dict] = None,
                 json: Optional[Dict] = None, timeout: int = 180, psk_param: str = "system_psk",
                 raise_error: bool = True) -> Any: ...

    def get_service_api(self, service_name: str) -> 'SecureWebAPI': ...


class DatabaseAPI(Protocol):
    dbs: Dict[str, str]
    conn_string: str
    client: Any
    backend: Optional[Database]
    price: Optional[Database]
    payroll: Optional[Database]
    auth: Optional[Database]
    forecast: Optional[Database]
    ims: Optional[Database]
    valuation: Optional[Database]

    def get_db(self, db_name: str) -> Optional[Database]: ...


class WebAPIWrapper(Protocol):
    pass


class GravitateAPIClient(Protocol):
    system: System
    web: SecureWebAPI
    db: DatabaseAPI
    client_id: Optional[str]
    client_secret: Optional[str]
    base_url: str
    url: str
    system_name: Optional[str]
    short_name: Optional[str]
    username: Optional[str]
    password: Optional[str]
    psk: Optional[str]
    connection_string: Optional[str]
    dbs: Optional[Dict[str, str]]
    api_map: Dict[Any, WebAPIWrapper]
    apis: WebAPIWrapper


class SDWebAPIWrapper(WebAPIWrapper):
    backend: GravitateAPIClient
    forecast: GravitateAPIClient
    valuation: GravitateAPIClient


class PEWebAPIWrapper(WebAPIWrapper):
    integration: GravitateAPIClient


class RitaWebAPIWrapper(WebAPIWrapper):
    backend: GravitateAPIClient


from typing import Optional

from pydantic import BaseModel, Field


class System(BaseModel):
    """A System is a specific software product. Systems may or may not belong to a company in the crossroads network."""
    name: str = Field(..., description="Name of the system")
    description: Optional[str] = Field(None, description="Details")
    is_erp: bool = Field(False, description="Does this sytem provide ERP functionality?")
    is_tms: bool = Field(False, description="Does this sytem provide TMS functionality?")
    is_active: bool = Field(True)


class Integration(BaseModel):
    """An Integration describes a runnable integration that can be used in the crossroads network."""
    name: str = Field(..., description="Name of the integration")
    description: Optional[str] = Field(None, description="Detailed description of the integration.")
    supersedes: Optional[str] = Field(None, description="ID of the integration that this integration supersedes. It is suggested to use this integration instead of the linked integration.")
    config_schema: dict = Field(default_factory=dict, description="A JSON schema for the configuration that this integration expects.")
    default_config: dict = Field(default_factory=dict, description="The default configuration values for this job.")
    is_active: bool = Field(True)


class Node(BaseModel):
    """A node is best thought of as a Company @ a System. So Caseys@Gravitate S+D, TTE@Gravitate S+D, Eagle@Telapoint are 3 nodes."""
    company_id: str = Field(..., description="Mongodb ID of the company")
    system_id: str = Field(..., description="Mongodb ID of the system")
    is_phantom: bool = Field(False, description="Should this be shown as a phantom node?")
    is_active: bool = Field(True)


class Connection(BaseModel):
    """A connection is an instance of an integration. It points between two Nodes, references an integration to run, and specifies the config for this integration."""
    name: str = Field(..., description="Name of the connection")
    description: Optional[str] = Field(None, description="Detailed description of the connection")
    origin_node_id: str = Field(..., description="ID of the origin node")
    target_node_id: str = Field(..., description="ID of the target node")
    integration_id: str = Field(..., description="ID of the integration")
    config: dict = Field(..., description="Config of this instance of the integration. This must conform to the schema defined by that integration.")
    is_active: bool = Field(True)


from pydantic import BaseModel, Field
from datetime import datetime, UTC, timedelta
from typing import List, Dict, Optional, Any
from enum import Enum

class ProcessReportBase(BaseModel):
    created_on: datetime = datetime.now(UTC)
    indexed_field: str
    included: Optional[List[Dict[str, str]]] = []
    config_id: Optional[str] = None
    snapshot_id: Optional[str] = None
    logs: Optional[Any] = None

    class Config:
        arbitrary_types_allowed = True

class ProcessReportV2Status(str, Enum):
    start = "start" # while start -> in progress = True
    stop = "stop" # means halt
    error = "error" #  full failure
    partial = "partial" # partial success

    @property
    def in_progress(self) -> bool:
        return self == ProcessReportV2Status.start

class ProcessReportFileReference(BaseModel):
    file_base_name: str = None
    created_on: Optional[datetime] = datetime.now(UTC)
    status: Optional[ProcessReportV2Status] = None


class UploadProcessReportFile(BaseModel):
    file_base_name: str
    content: str


class ProcessReportBaseV2(BaseModel):
    # This doesn't seem to conflict with Beanie's internal _id field, but it allows us to deserialize an _id coming
    # from rita-backend.
    id: Optional[str] = Field(None, frozen=True, alias="_id")
    trigger: str
    status: Optional[ProcessReportV2Status] = ProcessReportV2Status.start
    config_id: Optional[str] = None
    updated_on: Optional[datetime] = datetime.now(UTC)
    created_on: Optional[datetime] = datetime.now(UTC)
    logs: Optional[List[ProcessReportFileReference]] = []
    included_files: Optional[List[ProcessReportFileReference]] = []

    @property
    def time_delta(self) -> timedelta:
        return self.updated_on - self.created_on

    class Config:
        arbitrary_types_allowed = True


class CreateReportV2(BaseModel):
    """
    Used for creating a process report with a Rita endpoint.
    :param bool alert_override: Whether to override (force) send an alert.
    """
    alert_override: bool = False
    trigger: str
    status: Optional[ProcessReportV2Status] = ProcessReportV2Status.start
    config_id: Optional[str] = None
    log: Optional[UploadProcessReportFile] = None
    included_files: Optional[list[UploadProcessReportFile]] = None


class UpdateReportV2(BaseModel):
    """Used for updating an already-created report with a Rita endpoint."""
    report_id: str
    alert_override: bool = False
    log: Optional[UploadProcessReportFile] = None
    included_files: Optional[list[UploadProcessReportFile]] = None
    status: ProcessReportV2Status = None


    @property
    def in_progress(self) -> bool:
        return self == ProcessReportV2Status.start


class ProcessReportResponseV2(ProcessReportBaseV2):
    logs: Optional[List[str]] = None


class NotificationChannel(str, Enum):
    email = "email"
    slack = "slack"


class NotificationStatus(str, Enum):
    sent = "sent"  # email has been sent
    pending = "pending"  # alert is yet to be sent
    passed = "pass"  # alert is not needed
    alert = "alert"  # alert the UI


class Notification(BaseModel):
    notification_status: NotificationStatus = NotificationStatus.pending
    updated_on: datetime = datetime.now(UTC)
    channel: NotificationChannel = NotificationChannel.email
    recipients: Optional[list] = None


class AlertBase(BaseModel):
    config_id: Optional[str] = None
    process_id: Optional[str] = None
    trigger: Optional[str] = None
    created_on: Optional[datetime] = datetime.now(UTC).replace(tzinfo=UTC)
    notification: Notification = Notification()



import asyncio
from datetime import datetime, timedelta
from pprint import pprint
from typing import Optional, Self

import httpx
from loguru import logger

from bb_integrations_lib.protocols.flat_file import TankReading, TankMonitorType
from bb_integrations_lib.secrets.credential_models import WarrenRogersCredential


class WarrenRogersClient(httpx.AsyncClient):
    def __init__(
            self,
            client_id: str,
            client_secret: str,
            api_key: str,
            company_id: str,
            token_endpoint: str,
            base_url: str,
            timeout: float = 180.0
    ):
        super().__init__(base_url=base_url, timeout=timeout)
        self.client_id = client_id
        self.client_secret = client_secret
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key
        self.company_id = company_id
        self.token_endpoint = token_endpoint
        self.page_size = 100
        self.oauth_token = None
        # Expect to refresh token 30s before actual expiry, in case of time desync between us and server
        self.pre_expiry_refresh = timedelta(seconds=30)

    @classmethod
    def from_credential(cls, credential: WarrenRogersCredential) -> Self:
        return cls(
            **credential.model_dump(exclude={"type_tag"}),
        )

    def __repr__(self):
        return "Warren Rogers API Client"

    @property
    async def access_token(self):
        await self.ensure_token()
        return self.oauth_token["access_token"]

    async def _refresh_token(self):
        now = datetime.now()
        res = await self.post(self.token_endpoint,
                              auth=(self.client_id, self.client_secret),
                              headers={"Content-Type": "application/x-www-form-urlencoded"},
                              params="grant_type=client_credentials")
        res.raise_for_status()
        self.oauth_token = res.json()
        self.oauth_token["expires_at"] = now + timedelta(seconds=self.oauth_token["expires_in"])

    def is_token_expired(self):
        return datetime.now() > self.oauth_token["expires_at"] + self.pre_expiry_refresh

    async def ensure_token(self):
        # First run or expired token? Then refresh
        if not self.oauth_token or self.is_token_expired():
            await self._refresh_token()

    def _build_url(self, endpoint: str, api_ver: str = "v1") -> str:
        return f"api/{api_ver}/companies/{self.company_id}/{endpoint}"

    async def _get(self, url: str, params=None):
        headers = {
            "Accept": "application/json",
            "x-api-key": self.api_key,
            "wra-auth": await self.access_token
        }
        return await self.get(url, headers=headers, params=params)

    async def _paginated_get(self, url: str):
        pages = []
        page_no = 1

        first_page = (await self._get(url)).json()
        total_results = first_page["totalResults"]
        pages.append(first_page)
        entries_retrieved = first_page["size"]

        while entries_retrieved < total_results:
            page_no += 1
            page = (await self._get(url, params={"page": page_no})).json()
            pages.append(page)
            entries_retrieved += page["size"]

        return pages

    async def _continuation_get(self, url: str, starting_token: Optional[str] = None) -> tuple[str, list[dict]]:
        """
        Get data from a paginated stream that uses continuation tokens, like the /deliveries endpoint.
        If a starting continuation token is provided, it will be passed to the API; otherwise, the default API behavior
        will be used. In /deliveries case, this is retrieving the last hour of data, for example.

        :param url: The URL of the endpoint.
        :param starting_token: An optional continuation token to resume the stream from.
        :return: A tuple; item 1 is the continuation token from the API, and item 2 is the list of pages of the API
            response.
        """
        pages = []
        page_n = 1
        starting = True
        con_token = starting_token
        while starting or con_token:
            logger.debug(f"Getting page {page_n}, continuation token {con_token}")
            resp = await self._get(url=url, params={"continuationToken": con_token} if con_token else {})
            resp.raise_for_status()
            data = resp.json()
            con_token = data.get("continuationToken")
            pages.append(resp.json())
            if not data.get("truncated"):
                break
            starting = False
            page_n += 1
        return con_token, pages

    async def get_inventory_levels(self) -> list[dict]:
        """
        Get "inventory levels" (tank readings) for all tanks. Does not parse any values, but deserializes the json to a
        dict.
        """
        raw = await self._paginated_get(self._build_url("inventory-levels"))
        # The data this endpoint provides is a list of pages. Each page has some metadata plus an "inventoryLevels"
        # field, which has the actual locations and their ATGs and readings. Flatten to one big list of inventoryLevels.
        return [x for xs in raw for x in xs["inventoryLevels"]]

    async def get_inventory_levels_at_loc(self, location: str) -> list[dict]:
        """Like get_inventory_levels, but for a specific location."""
        raw = await self._get(self._build_url(f"locations/{location}/inventory-levels"))
        return raw.json()

    def _parse_tankreadings(self, location: dict) -> list[TankReading]:
        """Parse inventory levels from the WR API and convert them to TankReadings."""
        trs: list[TankReading] = []
        # For each ATG at location
        for atg_level in location["atgLevels"]:
            # For each of the actual inventory readings on this ATG
            ils = atg_level["tankInventoryLevels"]
            for il in ils:
                # Specific inventory level reading - i.e., tank
                # Note that this does not take the units reported by the WR API into account - this appears to be
                # gallons, so far.
                trs.append(
                    TankReading(
                        date=il["readingDateTime"],
                        payload={},
                        store=location["locationUniqueId"],
                        tank=il["tankUniqueId"],
                        timezone=None,  # Not needed, ISO date str has timezone
                        volume=il["atgGrossVolume"],
                        monitor_type=TankMonitorType.bbd
                    )
                )
        return trs

    async def get_unmapped_tank_readings(self) -> list[TankReading]:
        """
        Get all tank readings available to the client, like get_inventory_levels, but parse the tank readings with
        _parse_tankreadings, returning them as a flat list. Note that the store and tank identifiers are returned
        straight from the API, unmapped.
        """
        raw = await self._paginated_get(self._build_url("inventory-levels"))
        # The data this endpoint provides is a list of pages. Each page has some metadata plus an "inventoryLevels"
        # field, which has the actual locations and their ATGs and readings. Flatten to one big list of inventoryLevels.
        inventory_levels = [x for xs in raw for x in xs["inventoryLevels"]]
        trs = []
        # For each store
        for location in inventory_levels:
            trs.extend(self._parse_tankreadings(location))
        return trs

    async def get_unmapped_tank_readings_loc(self, location: str):
        """Like get_unmapped_tank_readings, but for a specific location."""
        raw = await self._get(self._build_url(f"locations/{location}/inventory-levels"))
        return self._parse_tankreadings(raw.json())

    async def get_deliveries(self, starting_token: Optional[str] = None) -> tuple[str, list[dict]]:
        """
        Get a list of locations and their deliveries. If starting_token is provided, resume the data stream from that
        continuation token. May return duplicates / repeated data.

        :param starting_token: An optional continuation token to resume the stream from. If not provided, gets the last
            hour of deliveries.
        :return: A tuple; item 1 is the final continuation token from the API (for resuming the stream in the future),
            and item 2 is the list of locations and their deliveries.
        """
        next_ct, raw = await self._continuation_get(
            self._build_url("deliveries", api_ver="v3"),
            starting_token=starting_token)
        # Flatten the locationDeliveries lists from each page into one list of dicts of locations and their deliveries.
        return next_ct, [x for xs in raw for x in xs["locationDeliveries"]]


if __name__ == "__main__":
    async def main():
        client = WarrenRogersClient(
            client_id="",
            client_secret="",
            api_key="",
            company_id="LOVES_TRAVEL_STOPS_AND_COUNTRY_STORES",
            token_endpoint="https://auth.api.wr-cloud.com/oauth2/token",
            base_url="https://api.wr-cloud.com")

        inventory_levels = await client.get_inventory_levels()
        pprint(inventory_levels)

    asyncio.run(main())



from typing import Optional, Dict, Protocol, List


class DTNClient(Protocol):
    async def get_allocations(self) -> List[Dict]:
        """Get all allocation data"""

    async def get_data_in_group(self, url: str) -> List[Dict]:
        """Get data in group href"""



from typing import Optional, Union

from pydantic import BaseModel, Field

from bb_integrations_lib.models.enums import ProbeEventType
from bb_integrations_lib.models.probe.request_data import RequestData, CreateOrderFromSDOrder, UpdateStatusFromSD, \
    CancelOrderFromSD, SaveBOLsAndDropsFromSD, ErrorRequestData, MacropointLocationUpdate


class ProbeEvent(BaseModel):
    source_probe: str
    type: ProbeEventType
    timestamp: str  # Datetime in ISOformat, UTC
    data: Union[CreateOrderFromSDOrder, UpdateStatusFromSD, CancelOrderFromSD, SaveBOLsAndDropsFromSD, ErrorRequestData, MacropointLocationUpdate] = Field(discriminator="request_type")

    record_id: Optional[str] = Field(description="The DB ID of the relevant record picked up by the probe. Must be set for all Crossroads operations.", default=None)
    record_id_field: Optional[str] = Field(description="The DB field containing the record_id. Must be set for all Crossroads operations.", default=None)
    record_collection: Optional[str] = Field(description="The DB collection this record belongs to. Must be set for all Crossroads operations.", default=None)





from zipfile import BadZipFile

import json
import math
import pandas as pd

from bb_integrations_lib.mappers.rita_mapper import AsyncMappingProvider, RitaAPIMappingProvider
from bb_integrations_lib.util.utils import CustomJSONEncoder
from loguru import logger
from pydantic import BaseModel
from typing import AsyncGenerator, Union
from typing import Tuple, List, Dict, Any, TypeVar

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.config import FileConfig, ConfigAction
from bb_integrations_lib.protocols.pipelines import Step, ParserBase, Parser
from bb_integrations_lib.shared.exceptions import FileParsingError
from bb_integrations_lib.shared.model import FileConfigRawData, MappingMode, RawData

AnyParser = TypeVar("AnyParser", bound=ParserBase | Parser)


class FileConfigParserJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, BaseModel):
            return o.model_dump()
        return super().default(o)


class FileConfigParserV2(Step):
    def __init__(self, rita_client: GravitateRitaAPI, mapping_type: MappingMode = MappingMode.full,
                 parser: type[AnyParser] | None = None, parser_kwargs: dict | None = None, *args, **kwargs):
        """
        Parse an input file using a FileConfig, and optionally, a custom parser.

        :param rita_client: GravitateRitaAPI instance
        :param mapping_type: How the parser should use mappings while processing rows. Defaults to "full"
        :param parser: Custom parser to pass translated rows through. Can be used to process many different types of
          files, such as tank readings or price tables.
        :param parser_kwargs: Additional keyword arguments to pass to the custom parser's init method.
        """
        super().__init__(*args, **kwargs)
        if parser:
            self.custom_parser: type[ParserBase] = parser
            self.custom_parser_kwargs = parser_kwargs or {}

        self.rita_client = rita_client
        self.mapping_type = mapping_type

    def describe(self) -> str:
        """Return a description of this step for logging."""
        return "Parse and translate file data using FileConfig"

    async def execute(self, rd: FileConfigRawData) -> Union[List[Dict], RawData]:
        """
        Execute file parsing and translation.
        Args:
            rd: FileConfigRawData containing file data and configuration

        Returns:
            List of translated records, or RawData if the custom parser returns RawData
        """
        file_name = rd.file_name or "<unknown filename>"
        self.pipeline_context.included_files[f"{file_name}"] = rd.data.read()
        translated_records, errors = self.get_translated_records(rd)
        mapping_provider = RitaAPIMappingProvider(rita_client=self.rita_client)
        if errors:
            logger.warning(f"Found {len(errors)} translation errors during parsing")
            self.pipeline_context.included_files["File Config Parser Step Translation Errors"] = json.dumps(errors)
        if not hasattr(self, "custom_parser"):
            self.pipeline_context.included_files["Parsed Results"] = json.dumps(translated_records)
            return translated_records
        else:
            logger.info(f"Using custom parser: {self.custom_parser.__name__}")
            parser = self.custom_parser(
                source_system=rd.file_config.source_system,
                file_name=rd.file_name,
                **self.custom_parser_kwargs
            )
            result = parser.parse(translated_records, self.mapping_type)
            if isinstance(result, AsyncGenerator):
                parser_results = [item async for item in result]
            else:
                parser_results = await result

            if isinstance(parser_results, RawData):
                logger.info(f"Parser returned RawData: {parser_results.file_name}")
                return parser_results

            self.pipeline_context.included_files[f"{self.custom_parser.__name__} Results"] = json.dumps(
                parser_results, cls=CustomJSONEncoder
            )
            return parser_results

    def _clean_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Clean and normalize dataframe:
        - Strip whitespace from column names
        - Strip whitespace from all string values
        - Remove duplicate rows

        Args:
            df: Input DataFrame to clean

        Returns:
            Cleaned DataFrame
        """
        df = df.rename(columns=lambda x: x.strip() if isinstance(x, str) else x)
        df = df.map(lambda x: x.strip() if isinstance(x, str) else x)
        df = df.drop_duplicates()
        return df

    def get_records(self, rd: FileConfigRawData) -> List[Dict]:
        """
        Parse file data into records based on file extension configuration.

        Supports multiple file formats:
        - csv: Standard CSV with headers
        - csv1: CSV with first row skipped
        - csv_headless: CSV without headers (generates col 1, col 2, etc.)
        - xls/xlsx: Excel files (tries openpyxl first, falls back to xlrd)
        - html: HTML tables

        Args:
            rd: FileConfigRawData containing file data and configuration

        Returns:
            List of dictionaries, each representing a row from the file

        Raises:
            FileParsingError: If file parsing fails for any reason
        """
        try:
            if hasattr(rd.data, 'seek'):
                rd.data.seek(0)

            match rd.file_config.file_extension:
                case "csv1":
                    # Skips first row
                    temp_df = pd.read_csv(
                        rd.data,
                        index_col=False,
                        dtype=str,
                        skiprows=1,
                        keep_default_na=False
                    )

                case "csv":
                    # Regular CSV
                    temp_df = pd.read_csv(
                        rd.data,
                        index_col=False,
                        dtype=str,
                        keep_default_na=False
                    )
                case "csv_headless":
                    # CSV missing headers
                    temp_df = pd.read_csv(
                        rd.data,
                        index_col=False,
                        dtype=str,
                        header=None,
                        keep_default_na=False
                    )
                    temp_df.columns = [f"col {i + 1}" for i in range(temp_df.shape[1])]
                case "fixed_width":
                    # Fixed with text file with headers
                    temp_df = pd.read_fwf(
                        rd.data,
                        sep=r'\s+',
                        dtype=str,
                        keep_default_na=False,
                    )
                case "fixed_width_headless":
                    # Fixed with text file without headers
                    temp_df = pd.read_fwf(
                        rd.data,
                        sep=r'\s+',
                        header=None,
                        dtype=str,
                        keep_default_na=False,
                    )
                    temp_df.columns = [f"col {i + 1}" for i in range(temp_df.shape[1])]
                case "xls" | "xlsx":
                    # Excel or XLSX
                    try:
                        temp_df = pd.read_excel(
                            rd.data,
                            engine="openpyxl",
                            dtype=str,
                            keep_default_na=False
                        )
                    except (OSError, BadZipFile):
                        temp_df = pd.read_excel(
                            rd.data,
                            engine="xlrd",
                            dtype=str,
                            keep_default_na=False
                        )

                case "html":
                    # HTML File
                    data = pd.read_html(rd.data, header=0, keep_default_na=False)
                    temp_df = pd.concat(data)
                    temp_df = temp_df.astype(str)

                case "override_header":
                    # Case where header is useless and we need to enumerate instead
                    # Should never need to use this
                    temp_df = pd.read_csv(
                        rd.data,
                        index_col=False,
                        dtype=str,
                        skiprows=1,
                        keep_default_na=False
                    )
                    temp_df.columns = [f"col {i + 1}" for i in range(temp_df.shape[1])]

                case _:
                    raise ValueError(
                        f"File extension '{rd.file_config.file_extension}' is not supported"
                    )

            temp_df = self._clean_dataframe(temp_df)
            records = temp_df.to_dict(orient="records")
            return records

        except Exception as e:
            msg = f"Failed to parse file with extension '{rd.file_config.file_extension}': {e}"
            logger.error(msg)
            raise FileParsingError(msg) from e

    def get_translated_records(self, rd: FileConfigRawData) -> Tuple[List[Dict], List[Dict]]:
        """
        Parse and translate file records using the file configuration.
        
        Extracts records from the file and applies column transformations
        defined in the file configuration. Collects translation errors
        without stopping processing.
        
        Args:
            rd: FileConfigRawData containing file data and configuration

        Returns:
            Tuple of (successfully_translated_records, translation_errors)
            - successfully_translated_records: List of translated record dictionaries
            - translation_errors: List of error dictionaries with 'row' and 'error' keys
        """
        translated_records = []
        translation_errors = []
        records = self.get_records(rd)
        for row in records:
            try:
                translated = FileConfigParserV2.translate_row(rd.file_config, row)
            except Exception as e:
                logger.warning(f"Translation failed for record {row}: {e}")
                translation_errors.append({
                    "row": row,
                    "error": str(e)
                })
                continue
            translated_records.append(translated)
        return translated_records, translation_errors

    @staticmethod
    def translate_row(file_config: FileConfig, row: dict) -> Dict:
        """
        Transform a single row based on file configuration column mappings.

        Applies various transformation actions to row data:
        - concat: Concatenate multiple columns/literals
        - parse_date: Parse date strings using pandas
        - concat_date: Concatenate multiple columns/literals with spaces, then parse as date
        - add: Add literal values
        - remove_leading_zeros/remove_trailing_zeros: Strip zeros
        - wesroc_volume_formula: Calculate Wesroc volume using the formula: (val1 * val2) / 100
        - default: Direct column mapping

        Sets failed column transformations to None and logs warnings.

        Args:
            file_config: FileConfig containing column transformation rules
            row: Dictionary representing a single data row

        Returns:
            Transformed row dictionary with mapped column names
        """
        output_row = {}
        for column in file_config.cols:
            try:
                if len(column.file_columns) == 0:
                    output_row[column.column_name] = None
                elif column.action == ConfigAction.concat:
                    concatenated = ""
                    for entry in column.file_columns:
                        stripped_entry = entry.strip()
                        if stripped_entry in row:
                            value = row[stripped_entry]
                            if value is None or (isinstance(value, float) and math.isnan(value) or pd.isna(value)):
                                concatenated += ""
                            else:
                                concatenated += str(value)
                        else:
                            # entry is not in col, concat it literally, don't strip it to avoid issues with intentional spaces
                            concatenated += str(entry)
                    output_row[column.column_name] = concatenated
                elif column.action == ConfigAction.parse_date:
                    if column.file_columns[0] not in row:
                        raise KeyError(f"Column '{column.file_columns[0]}' not found in row")
                    try:
                        output_row[column.column_name] = FileConfigParserV2._parse_datetime_to_string(
                            row[column.file_columns[0]],
                            column.format
                        )
                    except Exception as e:
                        raise ValueError(f"Failed to parse date from '{row[column.file_columns[0]]}': {e}")
                elif column.action == ConfigAction.concat_date:
                    concatenated_parts = []
                    for entry in column.file_columns:
                        stripped_entry = entry.strip()
                        if stripped_entry in row:
                            value = row[stripped_entry]
                            if value is None or (isinstance(value, float) and math.isnan(value) or pd.isna(value)):
                                continue
                            else:
                                concatenated_parts.append(str(value))
                        else:
                            concatenated_parts.append(str(entry))

                    concatenated = " ".join(concatenated_parts)
                    try:
                        output_row[column.column_name] = FileConfigParserV2._parse_datetime_to_string(
                            concatenated,
                            column.format
                        )
                    except Exception as e:
                        raise ValueError(f"Failed to parse date from concatenated string '{concatenated}': {e}")
                elif column.action == ConfigAction.add:
                    output_row[column.column_name] = column.file_columns[0]
                elif column.action == ConfigAction.remove_leading_zeros:
                    if column.file_columns[0] not in row:
                        raise KeyError(f"Column '{column.file_columns[0]}' not found in row")
                    output_row[column.column_name] = FileConfigParserV2.strip_leading_zeroes(
                        str(row[column.file_columns[0]]))
                elif column.action == ConfigAction.remove_trailing_zeros:
                    if column.file_columns[0] not in row:
                        raise KeyError(f"Column '{column.file_columns[0]}' not found in row")
                    output_row[column.column_name] = FileConfigParserV2.strip_trailing_zeroes(
                        str(row[column.file_columns[0]]))
                elif column.action == ConfigAction.wesroc_volume_formula:
                    if len(column.file_columns) != 2:
                        raise ValueError(
                            f"Wesroc volume formula action requires exactly 2 columns, got {len(column.file_columns)}")
                    if column.file_columns[0] not in row:
                        raise KeyError(f"Column '{column.file_columns[0]}' not found in row")
                    if column.file_columns[1] not in row:
                        raise KeyError(f"Column '{column.file_columns[1]}' not found in row")
                    output_row[column.column_name] = FileConfigParserV2.calculate_wesroc_volume(
                        row[column.file_columns[0]],
                        row[column.file_columns[1]],
                        column.file_columns[0],
                        column.file_columns[1]
                    )
                else:
                    if column.file_columns[0] not in row:
                        raise KeyError(f"Column '{column.file_columns[0]}' not found in row")
                    output_row[column.column_name] = str(row[column.file_columns[0]])
            except Exception as e:
                logger.warning(f"Failed to translate column '{column.column_name}': {e}")
                output_row[column.column_name] = None
        return output_row

    @staticmethod
    def _parse_datetime_to_string(value: str, date_format: str | None) -> str:
        """
        Parse a datetime value and convert it to a string.

        Args:
            value: The value to parse as a datetime
            date_format: Optional format string for parsing (None means auto-detect)

        Returns:
            String representation of the parsed datetime
        """
        return str(
            pd.to_datetime(
                value,
                format=date_format,  # None means auto-detect
                exact=False # Allow trailing data in timezone
            )
        )

    @staticmethod
    def strip_leading_zeroes(row: str) -> str:
        """
        Remove leading zeros from a string.
        
        Args:
            row: Input string to process
            
        Returns:
            String with leading zeros removed
        """
        return row.lstrip('0')

    @staticmethod
    def strip_trailing_zeroes(row: str) -> str:
        """
        Remove trailing zeros from a string.

        Args:
            row: Input string to process

        Returns:
            String with trailing zeros removed
        """
        return row.rstrip('0')

    @staticmethod
    def calculate_wesroc_volume(val1: Any, val2: Any, col1_name: str, col2_name: str) -> float:
        """
        Calculate Wesroc volume using the formula: (val1 * val2) / 100

        Args:
            val1: First value (typically quantity or percentage)
            val2: Second value (typically quantity or percentage)
            col1_name: Name of first column (for error messages)
            col2_name: Name of second column (for error messages)

        Returns:
            Calculated volume as float

        Raises:
            ValueError: If values are None/NaN or cannot be converted to float
        """
        if val1 is None or (isinstance(val1, float) and math.isnan(val1)) or pd.isna(val1):
            raise ValueError(f"Column '{col1_name}' contains null/NaN value")

        if val2 is None or (isinstance(val2, float) and math.isnan(val2)) or pd.isna(val2):
            raise ValueError(f"Column '{col2_name}' contains null/NaN value")

        try:
            return (float(val1) * float(val2)) / 100
        except (ValueError, TypeError) as e:
            raise ValueError(f"Failed to calculate Wesroc volume from '{val1}' and '{val2}': {e}")



from io import TextIOWrapper
from loguru import logger
from typing import Dict, AsyncIterable

from bb_integrations_lib.models.dtn_supplier_invoice import DTNSupplierInvoice, Parser
from bb_integrations_lib.protocols.pipelines import GeneratorStep
from bb_integrations_lib.shared.model import RawData


class ParseDTNInvoiceStep(GeneratorStep):
    def __init__(self, skip_failed: bool = True, *args, **kwargs) -> None:
        """
        Parse a DTN invoice and convert it into one or more modeled invoice items.

        :param skip_failed: Whether to skip any invoices in this file that fail to parse.
        """
        super().__init__(*args, **kwargs)
        self.skip_failed = skip_failed

    def describe(self) -> str:
        return "Parse a DTN supplier invoice file into one or more invoice models"

    async def generator(self, i: RawData) -> AsyncIterable[DTNSupplierInvoice]:
        try:
            p = Parser()
            tio = TextIOWrapper(i.data, encoding="utf-8")
            tio.seek(0)
            res = p.parse(tio)
            tio.detach() # Don't allow the CSV module to close the underlying BytesIO in case another step wants to use it
            for invoice in res:
                invoice.original_filename = i.file_name
                yield invoice
        except Exception as e:
            if self.skip_failed:
                logger.exception(f"Failed to process file {i.file_name}, skipping")
                return
            else:
                raise e



import json
import uuid
from typing import Any, List, Dict

from bb_integrations_lib.gravitate.sd_api import GravitateSDAPI
from bb_integrations_lib.shared.model import SDGetUnexportedOrdersRequest, SDGetUnexportedOrdersResponse
from bb_integrations_lib.protocols.pipelines import Step
from datetime import datetime, UTC, timedelta


class BBDGetOrdersToExportStep(Step):
    def __init__(self,
                 sd_client: GravitateSDAPI,
                 step_key: str | None = uuid.uuid4().hex,
                 target_date: datetime | None = None,
                 test_order_numbers: List[str] | None = None,
                 *args,**kwargs):
        super().__init__(*args,**kwargs)
        self.step_key = step_key
        self.sd_client = sd_client
        self.target_date = target_date
        self.test_override_order_numbers = test_order_numbers

    def describe(self):
        return "Get Orders to Export from SD"

    async def execute(self, _: Any = None) -> List[SDGetUnexportedOrdersResponse]:
        last_sync_date = self.last_sync_date()
        orders, serialized = await self.get_orders_since_last_check_in(last_sync_date)
        if self.pipeline_context is not None:
            self.pipeline_context.extra_data[self.step_key] = {
                "date_used_in_api_call": last_sync_date,
                "orders": serialized,
                "total_orders": len(serialized),
            }
            self.pipeline_context.extra_data["unexported_orders"] = orders
        return orders

    async def get_order_id(self, order_number: str) -> str | None:
        """Retrieve the order ID for a given order number."""
        response = await self.sd_client.get_orders(order_number=order_number)
        orders = response.json()
        if orders:
            return orders[0].get("order_id")
        return None

    async def create_test_orders(self) -> tuple[List[SDGetUnexportedOrdersResponse], List[Dict]]:
        """Create order response objects for test order numbers without calling get_unexported_orders."""
        orders_json = []
        for order_number in self.test_override_order_numbers:
            order_id = await self.get_order_id(order_number)
            orders_json.append({
                "order_id": order_id or "",
                "order_number": order_number,
                "export_status": "pending"
            })

        if self.pipeline_context is not None:
            self.pipeline_context.included_files[self.step_key] = json.dumps(orders_json)

        return [SDGetUnexportedOrdersResponse.model_validate(order) for order in orders_json], orders_json

    async def get_orders_since_last_check_in(self, last_check_in_date: datetime | None = None) -> tuple[List[
        SDGetUnexportedOrdersResponse], List[Dict]]:
        # If test order numbers provided, bypass API
        if self.test_override_order_numbers is not None:
            return await self.create_test_orders()

        # Normal path - call API
        req = SDGetUnexportedOrdersRequest(as_of=last_check_in_date)
        orders = await self.sd_client.get_unexported_orders(req=req)
        orders_json = orders.json()

        if self.pipeline_context is not None:
            self.pipeline_context.included_files[self.step_key] = json.dumps(orders_json)
        return [SDGetUnexportedOrdersResponse.model_validate(order) for order in orders_json] or [], orders_json

    def last_sync_date(self) -> datetime:
        if self.target_date is not None:
            return self.target_date
        if self.test_override_order_numbers is not None:
            return datetime.now(tz=UTC) - timedelta(days=45)
        return datetime.now(UTC)





from io import BytesIO

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.email import EmailData, EmailAttachment
from bb_integrations_lib.pipelines.steps.send_rita_email_step import SendRitaEmailStep
from bb_integrations_lib.shared.model import RawData


class SendAttachedInRitaEmailStep(SendRitaEmailStep):
    def __init__(self, rita_client: GravitateRitaAPI, to: str | list[str], html_content: str, subject: str,
                 timeout: float = 10.0, *args, **kwargs):
        """
        Send one or more RawData objects as an email attachment via RITA.

        :param rita_client: Instantiated RITA API client using an API key with email.send scope.
        :param to: Email address(es) to send the email to.
        :param html_content: HTML content (body) of the email.
        :param subject: Subject of the email.
        :param timeout: The maximum amount of time allowed to send the email. Large emails may take longer to send than
          the default.
        """
        super().__init__(rita_client=rita_client, timeout=timeout, *args, **kwargs)
        self.to = to
        self.html_content = html_content
        self.subject = subject

    def describe(self):
        return "Send email via RITA with RawData(s) from step input attached"

    async def execute(self, i: RawData | list[RawData]):
        if isinstance(i, RawData):
            i = [i]
        ed = EmailData(
            to=self.to,
            html_content=self.html_content,
            subject=self.subject,
            attachments=[
                EmailAttachment(
                    file_name=rd.file_name,
                    file_data=rd.data.getvalue() if type(rd.data) == BytesIO else rd.data,
                ) for rd in i
            ]
        )
        await super().execute(ed)



from typing import Any

from bb_integrations_lib.protocols.pipelines import Step, Input, Output


class NullStep(Step[Any, None]):
    """A step that performs no action. May be useful as the first step in a job pipeline."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def describe(self) -> str:
        return "Null Step"

    async def execute(self, i: Any) -> None:
        return None



import unittest
import re


def file_exact_match(file_name: str, sub_str: str, sep='_') -> bool:
    if not file_name or not sub_str:
        return False
    if not file_name.startswith(sub_str):
        return False
    remainder = file_name[len(sub_str):]
    if not remainder:
        return True
    if remainder.startswith('.'):
        return True
    if remainder.startswith(sep):
        date_part = remainder[1:]
        date_pattern = r'^(\d{4}-\d{2}-\d{2}|\d{2}-\d{2}-\d{4}|\d{8}|\d{4}_\d{2}_\d{2}|' + \
                       r'\d{2}_\d{2}_\d{4}|\d{4}\d{2}\d{2})(\.[a-zA-Z0-9]+)?$'
        return bool(re.match(date_pattern, date_part))

    return False


class TestFileExactMatch(unittest.TestCase):
    def test_examples_from_requirements(self):
        self.assertFalse(file_exact_match("Gross_contractrack.csv", "contractrack"))
        self.assertTrue(file_exact_match("contractrack.csv", "contractrack"))
        self.assertTrue(file_exact_match("contractrack_2025-04-16.csv", "contractrack"))
        self.assertFalse(file_exact_match("Gross_contractrack_2025-04-16.csv", "contractrack"))

    def test_valid_date_formats(self):
        self.assertTrue(file_exact_match("contractrack_20250416.csv", "contractrack"))
        self.assertTrue(file_exact_match("contractrack_2025-04-16.csv", "contractrack"))
        self.assertTrue(file_exact_match("contractrack_2025_04_16.csv", "contractrack"))

    def test_invalid_formats(self):
        # Invalid because there's additional text after the date
        self.assertFalse(file_exact_match("contractrack_2025-04-16_final.csv", "contractrack"))

        # Invalid because there's text instead of a date
        self.assertFalse(file_exact_match("contractrack_version2.csv", "contractrack"))

        # Invalid because substring isn't at the start
        self.assertFalse(file_exact_match("prefix_contractrack_2025-04-16.csv", "contractrack"))

    def test_different_separator(self):
        self.assertTrue(file_exact_match("contractrack-2025-04-16.csv", "contractrack", sep='-'))
        self.assertFalse(file_exact_match("Gross-contractrack-2025-04-16.csv", "contractrack", sep='-'))

    def test_edge_cases(self):
        self.assertTrue(file_exact_match("contractrack", "contractrack"))
        self.assertFalse(file_exact_match("contractrack_", "contractrack"))
        self.assertFalse(file_exact_match("contractrack_notadate.csv", "contractrack"))


if __name__ == "__main__":
    unittest.main()


import base64
from typing import Optional

from pydantic import BaseModel, model_serializer, computed_field


class EmailAttachment(BaseModel):
    file_name: str
    file_data_encoded: Optional[str] = None

    def __init__(self, file_name: str, file_data: Optional[bytes] = None, file_data_encoded: Optional[str] = None):
        super().__init__(file_name=file_name)
        self.file_name = file_name
        self.file_data_encoded = self._b64enc(file_data) if file_data is not None else file_data_encoded

    def _b64enc(self, v) -> str:
        return base64.b64encode(v).decode("utf-8")

    @computed_field
    def file_data(self) -> bytes:
        return base64.b64decode(self.file_data_encoded)

    @file_data.setter
    def file_data(self, new_file_data: bytes) -> None:
        self.file_data_encoded = self._b64enc(new_file_data)

    @model_serializer()
    def serialize_model(self):
        return {
            "file_name": self.file_name,
            "file_data_encoded": self.file_data_encoded
        }


class EmailData(BaseModel):
    to: str | list[str]
    html_content: str
    subject: str
    attachments: Optional[list[EmailAttachment]] = None



from datetime import datetime, UTC
from typing import Self

from httpx import Response

from bb_integrations_lib.gravitate.base_api import BaseAPI
from bb_integrations_lib.provider.api.platform_science.model import JobDefinition, LoadDefinition
from bb_integrations_lib.secrets.credential_models import PlatformScienceCredential


class PlatformScienceClient(BaseAPI):
    def __init__(self,
                 base_url: str,
                 client_id: str,
                 client_secret: str
                 ):
        super().__init__()
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self._token = None

    def __repr__(self):
        return "Platform Science API client"

    @classmethod
    def from_credential(cls, credential: PlatformScienceCredential) -> Self:
        return cls(
            base_url=credential.base_url,
            client_id=credential.client_id,
            client_secret=credential.client_secret,
        )

    async def _get_token(self):
        try:
            resp = await self.post(
                url=f"{self.base_url}oauth/token",
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                    "api-version": "2.0"
                },
                json={
                    "grant_type": "client_credentials",
                    "scope": "admin workflow",
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                }
            )
        except Exception:
            raise ValueError(f"Error getting token from {self.base_url}")

        try:
            self._token = resp.json()["access_token"]
            return self._token
        except Exception:
            raise ValueError(f"Token response invalid from {self.base_url} -> {resp.status_code}")

    async def _auth_request(self, method: str, **kwargs):
        if not self._token:
            await self._get_token()

        headers = kwargs.pop("headers", {})
        headers["Authorization"] = f"Bearer {self._token}"
        headers["api-version"] = "2.0"
        headers["Content-Type"] = "application/json"
        kwargs["headers"] = headers
        kwargs["url"] = f"{self.base_url}{kwargs.get('url', '')}"
        kwargs["timeout"] = kwargs.get("timeout", 90)

        response = await getattr(self, method)(**kwargs)

        if response.status_code == 401:
            await self._get_token()  # refresh token
            headers["authorization"] = f"Bearer {self._token}"
            kwargs["headers"] = headers
            response = await getattr(self, method)(**kwargs)

        return response

    async def auth_post(self, **kwargs) -> Response:
        return await self._auth_request("post", **kwargs)

    async def auth_get(self, **kwargs) -> Response:
        return await self._auth_request("get", **kwargs)

    async def auth_patch(self, **kwargs) -> Response:
        return await self._auth_request("patch", **kwargs)

    async def auth_put(self, **kwargs) -> Response:
        return await self._auth_request("put", **kwargs)

    async def auth_delete(self, **kwargs) -> Response:
        return await self._auth_request("delete", **kwargs)

    async def paginated_get(self, url: str, params: dict | None = None) -> list[dict]:
        params = params or {}
        first_page = await self.auth_get(url=url, params=params)
        if first_page.status_code != 200:
            raise RuntimeError(f"Error getting assets from {self.base_url}")
        body = first_page.json()
        data = body["data"]
        total_pages = body["meta"]["pagination"]["total_pages"]
        while body["meta"]["pagination"]["current_page"] < total_pages:
            params["page"] = body["meta"]["pagination"]["current_page"] + 1
            next_page = await self.auth_get(url="admin/assets", params=params)
            body = next_page.json()
            data.extend(body["data"])
        return data

    async def get_assets(self) -> list[dict]:
        return await self.paginated_get(url="admin/assets")

    async def get_drivers(self) -> list[dict]:
        return await self.paginated_get(url="admin/drivers")

    async def create_assets(self, req: list[dict]) -> Response:
        return await self.auth_post(url="admin/assets", json=req)

    async def get_amqp_connection(self, channel_name: str):
        vhost = ""
        conn_str = f"amqps://{self.client_id}:{self.client_secret}@amqp.pltsci.com:5671/{vhost}"

    async def create_workflow_job(self, driver_id: str, job_definition: JobDefinition) -> Response:
        return await self.auth_post(url=f"drivers/{driver_id}/jobs", json={
            "job": job_definition.model_dump(mode="json", exclude_unset=True)
        })

    async def update_workflow_job(self, driver_id: str, job_id: str, job_definition: JobDefinition) -> Response:
        return await self.auth_put(url=f"drivers/{driver_id}/jobs/{job_id}", json={
            "job": job_definition.model_dump(mode="json", exclude_unset=True)
        })

    async def complete_workflow_job(self, job_id: str, completed_at: datetime | None = None) -> Response:
        return await self.auth_patch(url=f"jobs/{job_id}/status", json={
            "status": "tms_completed",
            "completed_at": (completed_at or datetime.now(tz=UTC)).astimezone(tz=UTC).isoformat()
        })

    async def delete_workflow_job(self, job_id: str) -> Response:
        return await self.auth_delete(url=f"jobs/{job_id}/status")

    async def create_load(self, driver_id: str, load_definition: LoadDefinition) -> Response:
        return await self.auth_post(
            url=f"admin/drivers/{driver_id}/loads",
            json=load_definition.model_dump(mode="json", exclude_none=False)
        )



import httpx
import loguru
from bb_integrations_lib.provider.api.telapoint.model import CheckConnectionMethod, SoapEnvelope, OrderGetByOrderNumberMethod, TelapointNewOrder, OrderAdd

namespaces = {
    'soapenv': "http://schemas.xmlsoap.org/soap/envelope/",
    'tel': "http://schemas.datacontract.org/2004/07/TelaPoint.Api.TelaFuel.Models",
    'v2': "http://api.telapoint.com/TelaFuel/v2",
    'i': "http://www.w3.org/2001/XMLSchema-instance",
    'wsse': "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd",
    'wsu': "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
}


class TelapointClient(httpx.AsyncClient):
    def __init__(self, username: str, password: str):
        super().__init__(
            base_url="https://api.telapoint.com/APIv2/TelaFuelService.svc",
            headers={"Content-Type": "text/xml"}
        )
        self.base_url = "https://api.telapoint.com/APIv2/TelaFuelService.svc"
        self.username = username
        self.password = password

    async def check_connection(self):
        req = SoapEnvelope.build(self.username, self.password, CheckConnectionMethod)

        response = await super().post(
            url="",
            headers={"SOAPAction": "http://api.telapoint.com/TelaFuel/v2/ITelaFuelService/CheckConnection"},
            content=req.to_xml()
        )
        loguru.logger.debug(response.status_code)
        return response

    async def get_order_by_order_number(self, order_number: str):
        req = SoapEnvelope.build(self.username, self.password, OrderGetByOrderNumberMethod, order_number=order_number)

        response = await super().post(
            url="",
            headers={"SOAPAction": "http://api.telapoint.com/TelaFuel/v2/ITelaFuelService/OrderGetByOrderNumber"},
            content=req.to_xml()
        )
        loguru.logger.debug(response.status_code)
        return response

    async def create_order(self, new_order: TelapointNewOrder):
        req = SoapEnvelope.build_with_body(self.username, self.password, OrderAdd, OrderAdd(new_order=new_order))

        response = await super().post(
            url="",
            headers={"SOAPAction": "http://api.telapoint.com/TelaFuel/v2/ITelaFuelService/OrderAdd"},
            content=req.to_xml()
        )
        loguru.logger.debug(response.status_code)
        return response


if __name__ == "__main__":
    import asyncio

    async def main():
        tc = TelapointClient("<>", "<>")
        resp = await tc.check_connection()
        resp = await tc.get_order_by_order_number(order_number="51412828")
        pass

    asyncio.run(main())



import base64
from functools import lru_cache
import httpx
from httpx import Response
from bb_integrations_lib.provider.api.macropoint.model import LocationUpdateRequest


class MacropointClient(httpx.AsyncClient):
    def __init__(self, username, password):
        super().__init__()
        self.username = username
        self.password = password
        self._headers = {}

    @lru_cache(maxsize=1)
    async def _get_token(self):
        encoded = base64.standard_b64encode(f"{self.username}:{self.password}".encode()).decode()
        return f"Basic {encoded}"

    async def update_location(self, req: LocationUpdateRequest) -> Response:
        return await self.post(
            url="https://macropoint-lite.com/api/1.0/tms/data/location",
            headers={
                "Content-Type": "application/xml",
                "Authorization": await self._get_token(),
            },
            data=req.to_xml()
        )


import httpx

from bb_integrations_lib.provider.api.cargas.model import CreateWholesaleTicketRequest, CreateWholesaleFeeLineRequest, \
    CreateWholesaleLineRequest


class CargasClient(httpx.AsyncClient):
    def __init__(self,
                 api_key: str,
                 base_url: str = " https://cargase.aeroenergy.com/cargasenergy/API"):
        super().__init__(base_url=base_url)
        self.api_key = api_key

    @property
    def custom_headers(self) -> dict:
        return {
            'APIKey': self.api_key,
            'Content-Type': 'application/json'
        }

    async def create_wholesale_ticket(self, data: CreateWholesaleTicketRequest) -> dict:
        url = '/CreateWholesaleTicket'
        res = await self.post(
            url=url, headers=self.custom_headers, json=data.model_dump(exclude_none=True, mode="json")
        )
        res.raise_for_status()
        return res.json()

    async def create_wholesale_line(self, data: CreateWholesaleLineRequest) -> dict:
        url = '/CreateWholesaleLine'
        res = await self.post(
            url=url, headers=self.custom_headers, json=data.model_dump(exclude_none=True, mode="json")
        )
        res.raise_for_status()
        return res.json()

    async def create_wholesale_fee_line(self, data: CreateWholesaleFeeLineRequest) -> dict:
        url = '/CreateWholesaleFeeLine'
        res = await self.post(
            url=url, headers=self.custom_headers, json=data.model_dump(exclude_none=True, mode="json")
        )
        res.raise_for_status()
        return res.json()



import numpy as np
import pandas as pd
from bb_integrations_lib.protocols.pipelines import Parser
from typing import override, Dict, List
from pandas import DataFrame


class TankConfigsParser(Parser):
    def __init__(self, tenant_name: str, source_system: str | None = None):
        super().__init__(source_system, tenant_name)

    def __repr__(self) -> str:
        return "Tank configs parser"

    @override
    async def parse(self, data: List[Dict], mapping_type: str | None = None) -> DataFrame:
        tc_df = pd.DataFrame(data)
        dos_columns = ['store_number', 'product', 'daily_lifting_estimate', 'measured_inventory']
        tc_df = tc_df[dos_columns]

        tc_df = tc_df.groupby(['store_number', 'product'])[
            ['daily_lifting_estimate', 'measured_inventory']].sum().reset_index()
        tc_df['dos'] = np.where(
            (tc_df['measured_inventory'] == 0) | (tc_df['daily_lifting_estimate'] == 0),
            'N/A',
            tc_df['measured_inventory'] / tc_df['daily_lifting_estimate']
        )
        tc_df['dos'] = pd.to_numeric(tc_df['dos'], errors='coerce')
        tc_df['dos_bucket'] = np.where(
            tc_df['dos'] <= 2, '0-2',
            np.where(tc_df['dos'] <= 4, '2-4',
                     np.where(tc_df['dos'] <= 6, '4-6',
                              np.where(tc_df['dos'] <= 8, '6-8',
                                       np.where(tc_df['dos'] <= 10, '8-10',
                                                np.where(tc_df['dos'].isna(), 'N/A', '10+')))))
        )
        tc_df['product'] = tc_df['product'].apply(TankConfigsParser.normalize_product)
        tc_df['store_number'] = tc_df['store_number'].astype(str)

        return tc_df

    @staticmethod
    def normalize_product(val):
        try:
            return str(int(float(val)))
        except (ValueError, TypeError):
            return str(val).strip()


from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.rita.email import EmailData
from bb_integrations_lib.protocols.pipelines import Step


class SendRitaEmailStep(Step):
    def __init__(self, rita_client: GravitateRitaAPI, timeout: float = 10.0, raise_on_error: bool = True,
                 email_data_override: EmailData | None = None, *args, **kwargs):
        """
        Instantiate a pipeline step that sends an email via RITA.

        :param rita_client: Instantiated RITA API client using an API key with email.send scope.
        :param timeout: The maximum amount of time allowed to send the email. Large emails may take longer to send than
          the default.
        :param raise_on_error: Whether to raise an exception if the email send HTTP response indicates an error.
        :param email_data_override: Explicitly specified EmailData object to use instead of step input. This allows
          specifying a static email for either testing or notifications.
        """
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.timeout = timeout
        self.raise_on_error = raise_on_error
        self.ed_override = email_data_override

    def describe(self) -> str:
        return "Send email via RITA"

    async def execute(self, i: EmailData):
        if self.ed_override:
            resp = await self.rita_client.send_email(self.ed_override, self.timeout)
        else:
            resp = await self.rita_client.send_email(i, self.timeout)
        if self.raise_on_error:
            resp.raise_for_status()



from io import StringIO
import numpy as np
import pandas as pd
from bb_integrations_lib.protocols.pipelines import Parser
from typing import override
from pandas import DataFrame


class OrderBySiteProductParser(Parser):
    def __init__(self, tenant_name: str, source_system: str | None = None):
        super().__init__(source_system, tenant_name)

    def __repr__(self) -> str:
        return "Order by site and product parser"

    @override
    async def parse(self, data: str, mapping_type: str | None = None) -> DataFrame:
        orders = pd.read_csv(StringIO(data))
        orders_columns = ['order_number', 'site', 'site_name', 'component_product', 'finished_product',
                          'component_volume',
                          'contract', 'market']
        orders = orders[orders_columns]
        orders['contract_type'] = np.where(orders['contract'].isna(), 'rack', 'contract')
        grouped_sum = orders.groupby(['site', 'component_product', 'contract_type', 'market'])[
            'component_volume'].sum().reset_index()

        pivot_df = grouped_sum.pivot_table(
            index=['site', 'component_product'],
            values=['component_volume'],
            columns=['contract_type'],
            aggfunc='sum',
            fill_value=0
        )
        pivot_df.columns = ['_'.join(col).strip() if isinstance(col, tuple) else col for col in pivot_df.columns]
        pivot_df['component_volume_contract'] = pivot_df.get('component_volume_contract', 0)
        pivot_df['component_volume_rack'] = pivot_df.get('component_volume_rack', 0)
        pivot_df['total_product'] = pivot_df['component_volume_contract'] + pivot_df['component_volume_rack']
        pivot_df['pct_contract'] = pivot_df['component_volume_contract'] / pivot_df['total_product']
        pivot_df['pct_rack'] = pivot_df['component_volume_rack'] / pivot_df['total_product']
        pivot_df = pivot_df.reset_index()
        pivot_df['component_product'] = pivot_df['component_product'].apply(OrderBySiteProductParser.normalize_product)
        pivot_df['site'] = pivot_df['site'].astype(str)
        return pivot_df

    @staticmethod
    def normalize_product(val):
        try:
            return str(int(float(val)))
        except (ValueError, TypeError):
            return str(val).strip()


from time import sleep
from typing import Any, AsyncIterator

from loguru import logger

from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.models.pipeline_structs import NoPipelineData
from bb_integrations_lib.protocols.pipelines import GeneratorStep
from bb_integrations_lib.provider.imap.client import IMAPClient
from bb_integrations_lib.secrets import IMAPCredential
from bb_integrations_lib.shared.model import RawData, FileConfigRawData


class LoadIMAPAttachmentStep(GeneratorStep):
    def __init__(
            self,
            rita_client: GravitateRitaAPI,
            imap_client: IMAPClient,
            attachment_extension: str,
            to_email_address: str | None = None,
            from_email_address: str | None = None,
            delivered_to_email_address: str | None = None,
            retries: int = 3,
            email_subject: str | None = None,
            bucket_name: str | None = None,
            config_names: list[str] | None = None,
            raise_on_no_data: bool = False,
            *args, **kwargs

    ):
        """
        Load attachments from an IMAP folder / mailbox. The search criteriae are reductive; more filters are ANDed and
        will result in fewer results. If no filters are provided, the default is to search for all emails in the inbox.

        :param rita_client: The RITA client to retrieve fileconfigs with.
        :param imap_client: The IMAP account which owns the mailbox.
        :param attachment_extension: Sets the extension on the returned RawData.
        :param to_email_address: Matches emails sent TO this address, if provided.
        :param from_email_address: Matches emails sent FROM this address, if provided.
        :param delivered_to_email_address: Matches emails "Delivered-To" this address, if provided. Differs from the
          ``to_email_address`` argument - this matches on the Delivered-To mail header, which some email services set if
          they are automatically forwarding from their mailbox to ours.
        :param retries: How many times to retry a mail query that fails.
        :param email_subject: The subject of the email to search for.
        :param bucket_name: The config bucket that holds the config_names configs.
        :param config_names: The names of FileConfigs to use to search for.
        :param raise_on_no_data: Whether to raise an error when there are no matching emails.
        """
        super().__init__(*args, **kwargs)
        self.rita_client = rita_client
        self.imap_client = imap_client

        self.to_email_address = to_email_address
        self.from_email_address = from_email_address
        self.delivered_to_email_address = delivered_to_email_address
        self.email_subject = email_subject
        self.attachment_extension = attachment_extension
        self.retries = retries
        criteria_parts = ['UNSEEN']

        if self.from_email_address:
            criteria_parts.append(f'FROM {self.from_email_address}')

        if self.delivered_to_email_address:
            criteria_parts.append(f'HEADER Delivered-To {self.delivered_to_email_address}')
        if self.to_email_address:
            criteria_parts.append(f'TO {self.to_email_address}')

        if self.email_subject:
            if ' ' in self.email_subject:
                criteria_parts.append(f'SUBJECT "{self.email_subject}"')
            else:
                criteria_parts.append(f'SUBJECT {self.email_subject}')
        self.criteria = '(' + ' '.join(criteria_parts) + ')'

        self.bucket_name = bucket_name
        self.config_names = config_names

        self.raise_on_no_data = raise_on_no_data

    def describe(self) -> str:
        return "Get attachments from IMAP folder"

    async def generator(self, i: Any) -> AsyncIterator[RawData]:
        for config_name in self.config_names:
            async for result in self.load_using_config(config_name):
                yield result

    async def load_using_config(self, config_name: str) -> AsyncIterator[RawData]:
        file_config = (
            await self.rita_client.get_fileconfig_by_name(self.bucket_name, config_name)
        )[config_name]
        logger.info(f"Searching with criteria {self.criteria}")
        message_indexes = self.imap_client.search(self.criteria)
        logger.info(f"Found {len(message_indexes)} new emails in inbox meeting search criteria")
        if self.raise_on_no_data and not message_indexes:
            raise NoPipelineData("No new emails found in inbox meeting search criteria")
        for idx in message_indexes:
            logger.info(f"Fetching mail {idx}")
            for retry in range(self.retries):
                try:
                    message = self.imap_client.fetch(idx)
                    attachment_rd = self.imap_client.get_attachment_from_message(
                        message,
                        extension=self.attachment_extension,
                        return_rawdata=True
                    )
                    if attachment_rd is not None:
                        logger.info(f"Fetched attachment from email id: {idx}")
                        if file_config:
                            attachment_rd = FileConfigRawData(
                                file_name=attachment_rd.file_name,
                                data=attachment_rd.data,
                                file_config=file_config
                            )
                            self.pipeline_context.file_config = file_config
                        yield attachment_rd
                        break
                    else:
                        raise Exception("Returned attachment was None")
                except Exception as e:
                    logger.error(f"Error reading data from mail: {e}")
                    sleep(3)
                    self.imap_client.mark_unseen(idx)

    @staticmethod
    def _unplus_email(email_str: str) -> str:
        """
        Remove plus-addressing from an email to get the 'base' email account. If there is no plus, the original email
        string is returned. Note that in either case, the email must have an @ symbol in it.
        """
        split_at = email_str.split("@")
        if len(split_at) != 2:
            raise ValueError(f"Email address '{email_str}' is not in the expected format")
        first_half = split_at[0]
        has_plus = "+" in first_half
        if has_plus:
            unplussed = first_half[:first_half.rfind("+")]
            return f"{unplussed}@{split_at[1]}"
        return email_str


if __name__ == "__main__":
    async def main():
        step = LoadIMAPAttachmentStep(
            rita_client=GravitateRitaAPI(
                base_url="",
                client_id="",
                client_secret=""
            ),
            imap_client=IMAPClient(
                credentials=IMAPCredential(
                    host="",
                    port=993,
                    email_address="",
                    password=""
                )
            ),
            attachment_extension=".csv",
            bucket_name="/Inventory",
            config_names=["wawa"]
        )
        async for result in step.generator(None):
            print(result)


    import asyncio

    asyncio.run(main())



from datetime import datetime, timedelta
from typing import Iterable, Self

import pytz
from bb_integrations_lib.gravitate.rita_api import GravitateRitaAPI
from bb_integrations_lib.mappers.rita_mapper import RitaAPICachedMappingProvider
from bb_integrations_lib.pipelines.parsers.tank_reading_parser import TankReadingParser
from bb_integrations_lib.pipelines.steps.importing.load_imap_attachment_step import LoadIMAPAttachmentStep
from bb_integrations_lib.pipelines.steps.importing.sftp_file_config_step import SFTPFileConfigStep
from bb_integrations_lib.pipelines.steps.null_step import NullStep
from bb_integrations_lib.pipelines.steps.processing.archive_gcs_step import ArchiveGCSStep
from bb_integrations_lib.pipelines.steps.processing.archive_sftp_step import ArchiveSFTPStep
from bb_integrations_lib.pipelines.steps.processing.bbd_upload_tankreading_step import BBDUploadTankReadingStep
from bb_integrations_lib.pipelines.steps.processing.delete_sftp_step import DeleteSFTPStep
from bb_integrations_lib.pipelines.steps.processing.file_config_parser_step_v2 import FileConfigParserV2
from bb_integrations_lib.protocols.pipelines import JobPipeline
from bb_integrations_lib.secrets import SecretProvider
from bb_integrations_lib.secrets.factory import APIFactory
from bb_integrations_lib.shared.model import ImportTankReadings, ATGConfig, ConfigMode
from loguru import logger


class BBDImportTankReadingsPipeline(JobPipeline):
    @classmethod
    async def create(
            cls,
            config: ATGConfig,
            secret_provider: SecretProvider,
            rita_client: GravitateRitaAPI,
            bucket_name: str,
            config_name: str,
            tenant_name: str,
            send_reports: bool = True
    ):
        api_factory = APIFactory(secret_provider)

        steps = [{
            "id": "start",
            "parent_id": None,
            "step": NullStep({})
        }]
        if config.minutes_back is not None and config.timezone is not None:
            min_date = datetime.now(pytz.timezone(config.timezone)) - timedelta(minutes=config.minutes_back)
        else:
            min_date = None

        ftp_client = None
        if config.mode == "ftp":
            ftp_client = await api_factory.ftp(config.ftp_credentials)
            pull_step = SFTPFileConfigStep(
                rita_client=rita_client,
                ftp_client=ftp_client,
                bucket_name=bucket_name,
                config_name=config_name,
                mode=ConfigMode.ByName,
                match_mode=config.file_match_type,
                min_date=min_date or datetime.min
            )
        elif config.mode == "email":
            pull_step = LoadIMAPAttachmentStep(
                rita_client=rita_client,
                imap_client=await api_factory.imap(config.email_credentials),
                from_email_address=config.from_email_address,
                delivered_to_email_address=config.delivered_to_email_address,
                to_email_address=config.to_email_address,
                attachment_extension=config.attachment_extension or ".csv",
                bucket_name=bucket_name,
                config_names=[config_name],
                email_subject=config.email_subject
            )
        else:
            raise ValueError(f"Invalid mode: {config.mode}")

        mapping_provider = RitaAPICachedMappingProvider(rita_client)

        steps += [
            {
                "id": "pull",
                "parent_id": "start",
                "step": pull_step
            },
            {
                "id": f"parse_readings",
                "parent_id": "pull",
                "step": FileConfigParserV2(
                    rita_client=rita_client,
                    mapping_type=config.mapping_type,
                    parser=TankReadingParser,
                    parser_kwargs={
                        "included_payload": {
                            "source": f"{tenant_name} flat file tank readings"
                        },
                        "mapping_provider": mapping_provider,
                        "sd_client": await api_factory.sd(config.sd_credentials),
                    }
                )
            },
            {
                "id": f"bbd_upload",
                "parent_id": f"parse_readings",
                "step": BBDUploadTankReadingStep(
                    sd_client=await api_factory.sd(config.sd_credentials),
                )
            },
        ]
        if config.archive_gcs_bucket_path is not None and config.gcs_credentials is not None:
            steps.append({
                "id": f"archive_gcs",
                "parent_id": f"bbd_upload",
                "alt_input": "pull",
                "step": ArchiveGCSStep(
                    gcloud_storage=await api_factory.gcloud_storage(config.gcs_credentials),
                    bucket_path=config.archive_gcs_bucket_path,
                )
            })
        if config.archive_files:
            if ftp_client:
                steps.append({
                    "id": f"archive_ftp",
                    "parent_id": f"bbd_upload",
                    "alt_input": "pull",
                    "step": ArchiveSFTPStep(
                        ftp_client=ftp_client,
                    )
                })
            else:
                logger.warning("No FTP credentials provided, skipping FTP archive step")
        if config.delete_files:
            if ftp_client:
                steps.append({
                    "id": "delete_ftp",
                    "parent_id": f"bbd_upload",
                    "alt_input": "pull",
                    "step": DeleteSFTPStep(
                        ftp_client=ftp_client
                    )
                })
            else:
                logger.warning("No FTP credentials provided, skipping FTP delete step")
        config = await rita_client.get_config_by_name(bucket_path="/" + bucket_name, config_name=config_name)
        config_id = config[config_name].config_id
        return cls(
            steps,
            rita_client=rita_client,
            pipeline_name=f"Import tank readings - {config_name}",
            pipeline_config_id=config_id,
            secret_provider=secret_provider,
            send_reports=send_reports
        )


class MultiImportTankReadingsPipeline:
    def __init__(self, pipelines: Iterable[BBDImportTankReadingsPipeline]):
        self.pipelines = pipelines

    @classmethod
    async def create(cls, config_id: str, tenant_name: str, rita_client: GravitateRitaAPI,
                     secret_provider: SecretProvider, send_reports: bool = True) -> Self:
        main_config = await rita_client.get_config_by_id(config_id)
        [bucket] = [b for b in await rita_client.get_all_buckets() if b["_id"] == main_config.owning_bucket_id]

        import_config = ImportTankReadings.model_validate(main_config.config)
        pipelines = []
        for atg_config in import_config.configs:
            for config_name in atg_config.config_names:
                logger.info(f"Creating pipeline for config {config_name}")
                pipelines.append(await BBDImportTankReadingsPipeline.create(
                    config=atg_config,
                    secret_provider=secret_provider,
                    rita_client=rita_client,
                    bucket_name=bucket["name"],
                    config_name=config_name,
                    tenant_name=tenant_name,
                    send_reports=send_reports
                ))
        return cls(pipelines)

    async def run(self) -> None:
        if not self.pipelines:
            raise Exception("No pipelines to run - check config and try again")
        for pipeline in self.pipelines:
            try:
                logger.info(f"Starting pipeline '{pipeline.pipeline_name}'")
                await pipeline.execute()
                logger.info(f"Finished pipeline '{pipeline.pipeline_name}'")
                logger.info("-" * 40)
            except Exception as e:
                logger.exception(f"Failed to run pipeline '{pipeline.pipeline_name}'")



import json
import pprint
from json import JSONDecodeError

import httpx
from async_lru import alru_cache
from httpx import HTTPStatusError
from requests import Response

from bb_integrations_lib.models.rita.bucket import Bucket
from bb_integrations_lib.models.rita.crossroads_mapping import CrossroadsMapping, CrossroadsMappingResult, \
    MappingRequest
from bb_integrations_lib.models.rita.crossroads_monitoring import CrossroadsMappingError, CrossroadsError
from bb_integrations_lib.models.rita.reference_data import ReferenceDataMapping
from typing import Optional, Dict, Union, List, runtime_checkable, Protocol, Self

import loguru
from pydantic import ValidationError, BaseModel

from bb_integrations_lib.gravitate.base_api import BaseAPI
from bb_integrations_lib.secrets import RITACredential
from bb_integrations_lib.util.utils import CustomJSONEncoder
from bb_integrations_lib.models.probe.resume_token import ResumeToken
from bb_integrations_lib.models.rita.audit import CreateReportV2, ProcessReportBaseV2, UpdateReportV2
from bb_integrations_lib.models.rita.config import FileConfig, Config, GenericConfig, ConfigType, MaxSync
from bb_integrations_lib.models.rita.email import EmailData
from bb_integrations_lib.models.rita.issue import IssueBase, UpdateIssue
from bb_integrations_lib.models.rita.mapping import Map, UpdateMapItem
from bb_integrations_lib.models.rita.probe import ProbeConfig
from bb_integrations_lib.protocols.flat_file import TankReading


def catch_exceptions(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except httpx.HTTPStatusError as e:
            loguru.logger.error(f"HTTP status error: {e}")
            raise e
        except httpx.ConnectError as e:
            loguru.logger.error(f"Connection error: {e}")
            raise e
        except httpx.ReadTimeout as e:
            loguru.logger.error(f"Read timeout: {e}")
            raise e
        except Exception as e:
            loguru.logger.error(f"Unexpected error: {e}")
            raise e

    return wrapper


@runtime_checkable
class RitaBackendAPI(Protocol):
    async def get_mappings(self, source_system: Optional[str] = None, mapping_type: Optional[str] = None) -> list[Map]:
        """Gets all mappings by source_system and mapping type from RITA"""

    async def get_process_reports_sync(self, config_id: str) -> dict:
        """Gets latest sync date for a specific config"""

    async def get_file_configs(self, name: Optional[str] = None) -> dict[str, FileConfig]:
        """Gets File Configs from RITA"""

    async def create_process_report(self, new_process: CreateReportV2) -> ProcessReportBaseV2:
        """Creates a new process report in RITA"""

    async def get_config_by_id(self, id: str) -> Config:
        """Get a config by its id"""

    async def get_config_by_name(self, bucket_path: str, config_name: str) -> Dict[
        str, Union[FileConfig, ProbeConfig, GenericConfig]]:
        """Gets a config by its name"""


class GetMappings(BaseModel):
    source_system: Optional[str] = None


class GetConfigs(BaseModel):
    name: Optional[str] = None


class GravitateRitaAPI(BaseAPI):
    def __init__(
            self,
            client_id: str | None = None,
            client_secret: str | None = None,
            username: str | None = None,
            password: str | None = None,
            tenant: str | None = None,
            raise_errors: bool = True,
            base_url: str = "https://rita.gravitate.energy/api/"
    ):
        super().__init__(raise_errors)
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password
        self.tenant = tenant
        self._token = None

    async def _get_token(self):
        try:
            if self.username and self.password:
                resp = await self.post(
                    url=f"{self.base_url}auth/token",
                    data={"username": self.username, "password": self.password},
                )
            elif self.client_id and self.client_secret:
                resp = await self.post(
                    url=f"{self.base_url}auth/token",
                    data={"client_id": self.client_id, "client_secret": self.client_secret},
                )
            else:
                raise RuntimeError("Missing credentials for token request")
        except Exception:
            raise ValueError(f"Error Getting Token for {self.base_url}")

        try:
            self._token = resp.json()["access_token"]
            return self._token
        except Exception:
            raise ValueError(f"Could Not Get Token for {self.base_url} -> {resp.status_code}")

    async def _auth_req(self, method="POST", **kwargs):
        if not self._token:
            await self._get_token()

        headers = kwargs.pop("headers", {})
        headers["authorization"] = f"Bearer {self._token}"
        headers["X-Tenant-Name"] = self.tenant
        kwargs["headers"] = headers
        kwargs["url"] = f"{self.base_url}{kwargs.get("url", "")}"

        resp = await self.request(method, **kwargs)

        if resp.status_code == 401:
            await self._get_token()
            headers["authorization"] = f"Bearer {self._token}"
            kwargs["headers"] = headers
            resp = await self.request(method, **kwargs)

        if resp.status_code == 422:
            try:
                resp_content = pprint.pformat(resp.json())
            except JSONDecodeError:
                resp_content = resp.text
            raise HTTPStatusError(
                f"Bad request: \n{resp_content}",
                request=resp.request,
                response=resp
            )

        return resp

    @classmethod
    def from_credential(cls, credential: RITACredential) -> Self:
        return cls(
            base_url=credential.base_url,
            username=credential.username,
            password=credential.password,
            client_id=credential.client_id,
            client_secret=credential.client_secret,
            tenant=credential.tenant,
        )

    async def token_post(self, **kwargs):
        return await self._auth_req("POST", **kwargs)

    async def token_get(self, **kwargs):
        return await self._auth_req("GET", **kwargs)

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_file_configs(self, name: Optional[str] = None) -> dict[str, FileConfig]:
        try:
            _data = (GetConfigs(name=name).model_dump() if name else {})
            resp = await self.token_post(url="config/all", json=_data)
            resp.raise_for_status()
            return self.build_fileconfig_dict(resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while fetching configurations: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_config_by_name(self, bucket_path: str, config_name: str) -> Dict[
        str, Union[FileConfig, ProbeConfig, Dict]]:
        try:
            data = {"bucket_path": bucket_path, "name": config_name,
                    "bucket_name": bucket_path}  # Some older versions of RITA use bucket_name instead of bucket_path
            resp = await self.token_post(url="config/by_name", json=data)
            resp.raise_for_status()
            return self.build_config_dict([resp.json()])
        except Exception as e:
            loguru.logger.error(f"Error while fetching configs from bucket: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_all_buckets(self) -> list[dict]:
        resp = await self.token_post(url="bucket/all")
        resp.raise_for_status()
        return resp.json()

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_mappings(self, source_system: Optional[str] = None, mapping_type: Optional[str] = None) -> list[Map]:
        try:
            _data = (GetMappings(source_system=source_system).model_dump() if source_system else {})
            resp = await self.token_post(url="integration/mappings/all", json=_data)
            resp.raise_for_status()
            mappings: List[Map] = [Map.model_validate(obj) for obj in resp.json()]
            if mapping_type:
                mappings = list(filter(lambda m: m.type == mapping_type, mappings))
            return mappings
        except Exception as e:
            loguru.logger.error(f"Error while fetching mappings: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_mappings_by_source_system(self, source_system: str) -> list[Map]:
        def is_valid(obj) -> bool:
            try:
                map = Map.model_validate(obj)
                return map.is_active
            except ValidationError:
                return False

        try:
            resp = await self.token_post(url="mapping/by_source_system/", params={"source_system": source_system})
            resp.raise_for_status()
            mappings = [Map.model_validate(obj) for obj in resp.json() if is_valid(obj)]
            return mappings
        except Exception as e:
            loguru.logger.error(f"Error while fetching mappings: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_process_reports_sync(self, config_id: str, status: str = None):
        try:
            data = {
                "config_id": config_id,
                "status": status
            }
            resp = await self.token_post(url="process_report/latest_sync", params=data)
            resp.raise_for_status()
            return resp.json()
        except Exception as e:
            loguru.logger.error(f"Error while fetching process reports sync date: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_mappings_by_bucket_path(self, bucket_path: str) -> list[Map]:
        try:
            resp = await self.token_post(url="mapping/by_bucket_path", json=bucket_path)
            resp.raise_for_status()
            return [Map.model_validate(obj) for obj in resp.json()]
        except Exception as e:
            loguru.logger.error(f"Error while fetching mappings: {e}")
            raise e

    @catch_exceptions
    async def bulk_create_mappings(self, mappings: list[Map]) -> dict:
        resp = await self.token_post(
            url="integration/mappings/bulk_create",
            json=[m.model_dump(mode="json") for m in mappings],
        )
        resp.raise_for_status()
        return resp.json()

    @catch_exceptions
    async def bulk_update_mappings(self, mappings: list[UpdateMapItem]) -> dict:
        resp = await self.token_post(
            url="integration/mappings/bulk_update",
            json=[m.model_dump(mode="json") for m in mappings],
        )
        resp.raise_for_status()
        return resp.json()

    @catch_exceptions
    async def bulk_lock_mappings(self, mapping_ids: list[str], lock: bool = True) -> dict:
        resp = await self.token_post(
            url="integration/mappings/bulk_lock",
            json={"mapping_ids": mapping_ids, "lock": lock},
        )
        resp.raise_for_status()
        return resp.json()

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_mappings_by_source_system_v2(self, source_system: str) -> list[Map]:
        resp = await self.token_post(
            url="integration/mappings/by_source_system",
            json={"source_system": source_system},
        )
        resp.raise_for_status()
        return [Map.model_validate(obj) for obj in resp.json()]

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_mappings_by_group_id(self, group_id: str) -> list[Map]:
        resp = await self.token_post(
            url="integration/mappings/by_group_id",
            json={"group_id": group_id},
        )
        resp.raise_for_status()
        return [Map.model_validate(obj) for obj in resp.json()]

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_fileconfigs_from_bucket(self, bucket_path: str) -> dict[str, FileConfig]:
        try:
            resp = await self.token_post(url="config/from_bucket", json=bucket_path)
            resp.raise_for_status()
            return self.build_fileconfig_dict(resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while fetching configs from bucket: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_config_by_id(self, id: str) -> Config:
        try:
            resp = await self.token_post(url=f"config/{id}")
            resp.raise_for_status()
            return Config.model_validate(resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while fetching config: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_fileconfig_by_name(self, bucket_path: str, config_name: str) -> dict[str, FileConfig]:
        try:
            data = {"bucket_path": bucket_path, "name": config_name,
                    "bucket_name": bucket_path}  # Some older versions of RITA use bucket_name instead of bucket_path
            resp = await self.token_post(url="config/by_name", json=data)
            resp.raise_for_status()
            return self.build_fileconfig_dict([resp.json()])
        except Exception as e:
            loguru.logger.error(f"Error while fetching configs from bucket: {e}")
            raise e

    @catch_exceptions
    @alru_cache(maxsize=128)
    async def get_probeconfig_by_name(self, bucket_path: str, config_name: str) -> dict[str, ProbeConfig]:
        try:
            data = {"bucket_path": bucket_path, "name": config_name}
            resp = await self.token_post(url="config/by_name", json=data)
            resp.raise_for_status()
            configs = [Config.model_validate(resp.json())]
            return self.build_probeconfig_dict(configs)
        except Exception as e:
            loguru.logger.error(f"Error while fetching config '{config_name}' from bucket '{bucket_path}': {e}")
            raise e

    @catch_exceptions
    async def get_config_max_sync(self, config_id: str) -> Response:
        resp = await self.token_post(url="config/max_sync", params={"config_id": config_id})
        resp.raise_for_status()
        return resp

    @catch_exceptions
    async def update_config_max_sync(self, config_id: str, max_sync: MaxSync) -> Response:
        resp = await self.token_post(url="config/update_max_sync", params={"config_id": config_id},
                                     json=max_sync.model_dump(mode="json"))
        resp.raise_for_status()
        return resp

    @catch_exceptions
    async def get_resume_token(self, probe_id: str) -> ResumeToken | None:
        try:
            resp = await self.token_post(url=f"probe/get_resume_token/{probe_id}")
            resp.raise_for_status()
            resp = resp.json()
            if resp is not None:
                return ResumeToken.model_validate(resp)
            return None
        except Exception as e:
            loguru.logger.error(f"Error while fetching resume token: {e}")
            raise e

    @catch_exceptions
    async def get_available_tenants(self):
        try:
            resp = await self.token_post(url="/meta/tenant/available")
            resp.raise_for_status()
            return resp.json()
        except Exception as e:
            loguru.logger.error(f"Error while fetching available tenants: {e}")
            raise e

    @catch_exceptions
    async def set_resume_token(self, resume_token: ResumeToken):
        try:
            resp = await self.token_post(url=f"probe/set_resume_token", json=resume_token.model_dump())
            resp.raise_for_status()
            return resp.json()
        except Exception as e:
            loguru.logger.error(f"Error while setting resume token: {e}, resp={resp} {resp.content()}")
            raise e

    async def get_estick_wr(self, store_number: str, tank_number: str) -> TankReading:
        try:
            resp = await self.token_post(url="estick/wr", params={"store_number": store_number, "tank_id": tank_number})
            resp.raise_for_status()
            return TankReading(**resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while fetching Warren Rogers estick reading: {e}")
            raise e

    async def get_estick_wr_many(self, store_number: str, tank_numbers: list[str]) -> list[TankReading]:
        try:
            resp = await self.token_post(url="estick/wr/many",
                                         json={"store_number": store_number, "tank_ids": tank_numbers})
            resp.raise_for_status()
            return [TankReading(**reading) for reading in resp.json()]
        except Exception as e:
            loguru.logger.error(f"Error while fetching multiple Warren Rogers estick readings: {e}")
            raise e

    @catch_exceptions
    async def create_process_report(self, new_process: CreateReportV2) -> ProcessReportBaseV2:
        try:
            _data = new_process.model_dump(exclude_none=True)
            payload_json = json.dumps(_data, cls=CustomJSONEncoder)
            # 30s timeout - large files may take a long time to upload
            resp = await self.token_post(url="process_report/create", json=json.loads(payload_json), timeout=30.0)
            resp.raise_for_status()
            return ProcessReportBaseV2(**resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while creating process_report: {e}")
            raise e

    async def update_process_report(self, update_process: UpdateReportV2) -> ProcessReportBaseV2:
        try:
            _data = update_process.model_dump(exclude_none=True)
            payload_json = json.dumps(_data, cls=CustomJSONEncoder)
            resp = await self.token_post(url="process_report/update", json=json.loads(payload_json))
            resp.raise_for_status()
            return ProcessReportBaseV2(**resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while creating process_report: {e}")
            raise e

    async def record_many_issues(self, issues: List[IssueBase]):
        try:
            res = await self.token_post(url="issue/record_many",
                                        json=[issue.model_dump(mode="json") for issue in issues])
            res.raise_for_status()
            # Endpoint does not return anything other than the status code
        except Exception as e:
            loguru.logger.error(f"Error while recording issue(s): {e}")
            raise e

    async def record_issue(self, issue: IssueBase):
        await self.record_many_issues([issue])

    async def update_issue(self, issue: UpdateIssue) -> IssueBase:
        try:
            resp = await self.token_post(url="issue/update", content=issue.model_dump_json(exclude_unset=True))
            resp.raise_for_status()
            return IssueBase(**resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while updating issue {issue.key}: {e}")
            raise e

    async def get_issue(self, key: str) -> IssueBase:
        try:
            resp = await self.token_post(url=f"issue/get/{key}")
            resp.raise_for_status()
            return IssueBase(**resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while updating issue {key}: {e}")
            raise e

    async def get_issues_by_config_id(self, config_id: str) -> list[IssueBase]:
        try:
            resp = await self.token_post(url=f"issue/by_config_id/{config_id}")
            resp.raise_for_status()
            content = resp.json()
            return [IssueBase(**item) for item in content]
        except Exception as e:
            loguru.logger.error(f"Error while fetching issues by config id {config_id}: {e}")
            raise e

    async def call_ep(self, url: str, params: dict = None, json: dict = None, method: str = "POST") -> Response:
        if method == "POST":
            return await self.token_post(url=url, params=params, json=json)
        if json is not None:
            raise ValueError("JSON is not supported for GET requests")
        return await self.token_get(url=url, params=params)

    async def update_config_value(self, config_id: str, new_value: dict):
        try:
            config = await self.get_config_by_id(config_id)
            config.config = new_value
            update_req = json.loads(config.model_dump_json())
            update_req["_id"] = config_id
            resp = await self.token_post(url=f"config/update", json=update_req)
            resp.raise_for_status()
            return Config.model_validate(resp.json())
        except Exception as e:
            loguru.logger.error(f"Error while updating config {config_id}: {e}")
            raise e

    @catch_exceptions
    async def send_email(self, email: EmailData, timeout: float = 10.0) -> Response:
        """Sends an email via RITA. Default timeout is 10 seconds. Large emails may take longer to send."""
        return await self.token_post(url="integration/send_email", json=email.model_dump(), timeout=timeout)

    @alru_cache(maxsize=64)
    async def get_connection(self, connection_id: str) -> dict:
        return await self.token_get(url=f"crossroads/network/connections/{connection_id}")

    async def get_crossroads_mapping(self, req: MappingRequest) -> MappingRequest:
        try:
            resp = await self.token_post(url=f"crossroads_2/map", json=req.model_dump(mode="json"))
            if resp.status_code == 499:
                # There was an error while mapping. Parse the detail and re-raise the error.
                js = resp.json()["detail"]
                raise CrossroadsMappingError(**js)
            if resp.status_code == 500:
                raise ValueError(resp.content)
            content = resp.json()
            result = MappingRequest.model_validate(content)
            return result
        except Exception as e:
            loguru.logger.error(f"Unexpected error while fetching Crossroads mapping: {e}")
            raise e

    def build_single_fileconfig_dict(self, config: Config) -> FileConfig:
        """
        Converts a config to  File Config.
        """
        file_config = FileConfig.model_validate(config.config)
        file_config.config_id = config.id  # Add the config's ID to the fileconfig obj, in case it's needed later.
        return file_config

    def build_single_probeconfig_dict(self, config: Config) -> ProbeConfig:
        """
        converts a config into a ProbeConfig.
        """
        return ProbeConfig.model_validate(config.config)

    def build_config_dict(self, configs: List[Dict]) -> Dict[str, Union[FileConfig, ProbeConfig, GenericConfig]]:
        """Builds a config lkp by name and parses each config to its respective model"""
        output = {}
        for c in configs:
            config = Config.model_validate(c)
            if config.type == ConfigType.fileconfig:
                output[config.name] = self.build_single_fileconfig_dict(config)
            elif config.type == ConfigType.probeconfig:
                output[config.name] = self.build_single_probeconfig_dict(config)
            else:
                output[config.name] = GenericConfig(config_id=c.get("_id"), config=config.config)
        return output

    def build_probeconfig_dict(self, configs: List[Config]) -> Dict[str, ProbeConfig]:
        """
        Converts a list of configs of any type into a dict mapping config name to parsed ProbeConfig. Skips entries
        that are not ProbeConfigs, so the output dict may be smaller than the input list.
        """
        probe_configs = [c for c in configs if c.type == ConfigType.probeconfig]
        output = {}
        for config in probe_configs:
            parsed_probe_config = ProbeConfig.model_validate(config.config)
            output[config.name] = parsed_probe_config
        return output

    def build_fileconfig_dict(self, configs: list[dict]) -> dict[str, FileConfig]:
        """
        Converts a list of configs of any type into a dict mapping config name to parsed FileConfig. Skips entries
        that are not FileConfigs, so the output dict may be smaller than the input list.
        """
        output = {}
        for c in configs:
            config = Config.model_validate(c)
            if config.type == ConfigType.fileconfig:
                file_config = FileConfig.model_validate(config.config)
                file_config.config_id = c.get(
                    "_id")  # Add the config's ID to the fileconfig obj, in case it's needed later.
                output[config.name] = file_config
        return output
